""" .. _fabik_deploy_uwsgi:

fabik.deploy.uwsgi
~~~~~~~~~~~~~~~~~~~

封装远程 uwsgi 部署。
"""


from pathlib import Path

from fabric.connection import Connection
from invoke.exceptions import Exit
from fabik.deploy import Deploy


class UwsgiDeploy(Deploy):
    """ 使用 uWSGI 来部署服务
    """
    def __init__(self, fabik_conf: dict, work_dir: Path, conn: Connection, env_name: str | None=None):
        super().__init__(fabik_conf, work_dir, conn, env_name)

    def get_fifo_file(self):
        """ 使用 master-fifo 来管理进程
        http://uwsgi-docs-zh.readthedocs.io/zh_CN/latest/MasterFIFO.html
        """
        fifofile = self.get_remote_path('uwsgi.fifo')
        if self.remote_exists(fifofile):
            return fifofile
        return None

    def get_pid_file(self):
        """ 使用 pidfile 来判断进程是否启动
        """
        pidfile = self.get_remote_path('uwsgi.pid')
        if self.remote_exists(pidfile):
            return pidfile
        return None

    def get_uwsgi_exe(self):
        """ 获取 venv 中 uwsgi 的可执行文件绝对路径
        """
        uwsgi_exe = self.get_remote_path('venv/bin/uwsgi')
        if not self.remote_exists(uwsgi_exe):
            raise Exit('没有找到 uwsgi 可执行文件！请先执行 init_remote_venv')
        return uwsgi_exe

    def start(self):
        """ 启动服务进程
        """
        pidfile = self.get_pid_file()
        if pidfile is not None:
            raise Exit('进程不能重复启动！')
        self.conn.run(self.get_uwsgi_exe() + ' ' + self.get_remote_path('uwsgi.ini'))

    def stop(self):
        """ 停止 API 进程
        """
        fifofile = self.get_fifo_file()
        if fifofile is not None:
            self.conn.run('echo q > %s' % fifofile)
        pidfile = self.get_pid_file()
        # 删除 pidfile 以便下次启动
        if pidfile is not None:
            self.conn.run('rm %s' % pidfile)

    def reload(self):
        """ 优雅重载 API 进程
        """
        fifofile = self.get_fifo_file()
        if fifofile is None:
            raise Exit('进程还没有启动！')
        self.conn.run('echo r > %s' % fifofile)



""" .._fabik_util_encrypt:

fabik.util.encrypt
~~~~~~~~~~~~~~~~~~~

用于加密的封装
"""
import re
import base64

from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend

from fabik.error import EncryptError

ZERO_RE = re.compile(r'[\x00-\x1F]+$')


class AES_CBC(object):
    """使用 AES 进行对称加解密。

    :param key: 密钥。
    :param iv: iv。
    """

    def __init__(self, key, iv):
        self._cipher = Cipher(
            algorithms.AES(key), modes.CBC(iv), backend=default_backend()
        )

    def encrypt(self, plain_txt):
        return None

    def decrypt(self, cipher: str | bytes) -> str:
        decryptor = self._cipher.decryptor()
        if isinstance(cipher, str):
            cipher = cipher.encode()
        plain = decryptor.update(cipher) + decryptor.finalize()
        # 去掉尾部可能存在的 \x00-\x1F
        return ZERO_RE.sub('', plain.decode('utf-8'))

    def decrypt_b64(self, b64_txt) -> str:
        encrypted_bytes = base64.b64decode(b64_txt)
        return self.decrypt(encrypted_bytes)


class Encrypt(object):
    """使用 Fernet 模块进行对称加解密。

    :param key: 密钥。
    """

    @staticmethod
    def fernet_key() -> bytes:
        """生成一个 fernet 密钥。"""
        key = Fernet.generate_key()
        return key

    def __init__(self, key: str) -> None:
        self._factory = Fernet(key.encode())

    def encrypt(self, plain: str | bytes) -> bytes:
        """对提供的 plain_text 进行加密。"""
        if isinstance(plain, str):
            plain = plain.encode()
        return self._factory.encrypt(plain)

    def decrypt(self, cipher: str | bytes) -> bytes:
        """对提供的 cipher_text 进行解密。"""
        if isinstance(cipher, str):
            cipher = cipher.encode()
        try:
            txt = self._factory.decrypt(cipher)
            return txt
        except (InvalidToken, UnicodeDecodeError) as e:
            raise EncryptError(err_type=e, err_msg=f'Invalid Token: {cipher!s}, error: {e!s}')



""" .. _fabik_cmd:

fabik.cmd
----------------------------

fabik command line toolset
"""

__all__ = [
    "main_callback",
    "main_init",
    "conf_callback",
    "conf_tpl",
    "gen_requirements",
]

from enum import StrEnum
from ntpath import expanduser
import shutil
from pathlib import Path
from typing import Any, Callable

import jinja2
import typer
from typing import Annotated
import subprocess

import fabik
from fabik import tpl, util
from fabik.conf import (
    ConfigReplacer,
    FabikConfig,
    config_validator_name_workdir,
    config_validator_tpldir,
)
from fabric.connection import Connection
from fabik.error import (
    ConfigError,
    FabikError,
    PathError,
    echo_error,
    echo_info,
    echo_warning,
)


class UUIDType(StrEnum):
    UUID1 = "uuid1"
    UUID4 = "uuid4"


class DeployClassName(StrEnum):
    GUNICORN = "gunicorn"
    uWSGI = "uwsgi"


class GlobalState:
    cwd: Path
    conf_file: Path
    env: str | None = None
    force: bool = False
    fabic_config: FabikConfig | None = None
    _config_validators: list[Callable] = []  # 存储自定义验证器函数

    deploy_conn: "Deploy" = None  # type: ignore # noqa: F821

    @property
    def conf_data(self) -> dict[str, Any]:
        if self.fabic_config is None:
            return {}
        return self.fabic_config.root_data or {}

    def register_config_validator(self, validator_func: Callable) -> None:
        """注册一个配置验证函数，用于验证配置数据。

        验证函数应接受 FabikConfig 对象作为参数，返回 bool 值表示验证结果。
        如果验证失败，验证函数应自行处理错误提示。

        :param validator_func: 验证函数
        """
        if callable(validator_func) and validator_func not in self._config_validators:
            self._config_validators.append(validator_func)

    def _check_conf_data(self) -> bool:
        """执行所有已注册的配置验证器"""
        if not self.fabic_config:
            return False

        # 运行所有注册的验证器
        for validator in self._config_validators:
            try:
                if not validator(self.fabic_config):
                    return False
            except Exception as e:
                echo_error(f"Config validation error: {str(e)}")
                raise typer.Abort()

        return True

    def check_work_dir_or_use_cwd(self) -> Path:
        """如果提供了配置文件，检测 work_dir 并返回。否则使用 cwd 作为 work_dir。"""
        try:
            conf = FabikConfig(self.cwd, cfg=self.conf_file)
            conf.load_root_data()
            work_dir = Path(conf.getcfg("PATH", "work_dir"))
            if not work_dir.is_absolute():
                echo_warning(f"{work_dir} is not a absolute path.")
                raise typer.Exit()
            return work_dir
        except (PathError, ConfigError):
            pass
        return self.cwd

    def load_conf_data(
        self,
        check: bool = False,
        file_not_found_err_msg: str = 'Please call "fabik init" to generate a "fabik.toml" file.',
    ) -> dict[str, Any]:
        try:
            self.fabic_config = FabikConfig(self.cwd, cfg=self.conf_file)
            self.fabic_config.load_root_data()
            if check:
                self._check_conf_data()
            return self.conf_data
        except PathError as e:
            if isinstance(e.err_type, FileNotFoundError):
                echo_error(file_not_found_err_msg)
            else:
                echo_error(e.err_msg)
            raise typer.Abort()
        except ConfigError as e:
            echo_error(e.err_msg)
            raise typer.Abort()
        except FabikError as e:
            echo_error(e.err_msg)
            raise typer.Abort()
        except Exception as e:
            echo_error(e)
            raise typer.Abort()

    def write_config_file(
        self,
        tpl_name: str,
        /,
        tpl_dir: Path | None = None,
        target_postfix: str = "",
    ) -> None:
        """写入配置文件

        :param target_postfix: 配置文件的后缀
        """
        try:
            replacer = ConfigReplacer(
                global_state.conf_data,
                work_dir=global_state.cwd,  # type: ignore
                tpl_dir=tpl_dir,
                env_name=global_state.env,
            )
            target, final_target = replacer.set_writer(
                tpl_name, global_state.force, target_postfix
            )

            # 写入文件
            if replacer.writer:
                replacer.writer.write_file(global_state.force)
        except FabikError as e:
            echo_error(e.err_msg)
            raise typer.Abort()
        except Exception as e:
            echo_error(str(e))
            raise typer.Abort()

    def copy_tpl_file(self, keyname: str, filename: str, rename: bool = False):
        """复制文件到目标文件夹。

        :param rename: 若文件已存在是否重命名
        """
        split_path = keyname.split("/")

        if self.fabic_config is None:
            echo_warning("Please perform GlobalState.load_conf_data first.")
            raise typer.Exit()

        config_validator_tpldir(self.fabic_config)
        # 源文件夹
        srcfiledir = Path(self.fabic_config.getcfg("PATH", "tpl_dir"))
        # 目标文件夹
        dstfiledir = self.cwd
        while len(split_path) > 1:
            # 检测是否拥有中间文件夹，若有就创建它
            dstfiledir = dstfiledir.joinpath(split_path[0])  # type: ignore
            srcfiledir = srcfiledir.joinpath(split_path[0])
            if not dstfiledir.exists():
                dstfiledir.mkdir()
            split_path = split_path[1:]

        srcfile = srcfiledir / filename
        dstfile = dstfiledir / filename  # type: ignore

        if dstfile.exists():
            if self.force:
                shutil.copyfile(srcfile, dstfile)
                echo_info(f"覆盖 [red]{srcfile}[/] 到 [red]{dstfile}[/]")
            elif rename:
                dstbak = dstfile.parent.joinpath(dstfile.name + ".bak")
                if dstbak.exists():
                    echo_warning(
                        f"备份文件 [red]{dstbak}[/] 已存在！请先删除备份文件。"
                    )
                else:
                    shutil.move(dstfile, dstbak)
                    echo_info(f"备份文件 [yellow]{dstfile}[/] 到 [yellow]{dstbak}[/]")
                    shutil.copyfile(srcfile, dstfile)
                    echo_info(f"复制 {srcfile} 到 {dstfile}")
            else:
                echo_warning(f"文件 [red]{dstfile}[/] 已存在！")
        else:
            shutil.copyfile(srcfile, dstfile)
            echo_info(f"复制 [red]{srcfile}[/] 到 [red]{dstfile}[/]！")

    def build_deploy_conn(self, deploy_class: type["Deploy"]) -> "Deploy":  # type: ignore  # noqa: F821
        """创建一个远程部署连接。"""
        try:
            # 确保配置已加载
            if self.fabic_config is None:
                self.load_conf_data(check=True)

            # 使用已加载的配置
            replacer = ConfigReplacer(self.conf_data, self.cwd, env_name=self.env)  # type: ignore
            # 从 fabik.toml 配置中获取服务器地址
            fabric_conf = replacer.get_tpl_value("FABRIC", merge=True)
            pye_conf = replacer.get_tpl_value("PYE", merge=True)

            # 确保 fabric_conf 是一个字典
            if (
                not isinstance(fabric_conf, dict)
                or "host" not in fabric_conf
                or "user" not in fabric_conf
            ):
                raise ConfigError(
                    err_type=ValueError(),
                    err_msg="FABRIC configuration must contain 'host' and 'user' parameter",
                )

            if pye_conf is None:
                raise ConfigError(
                    err_type=ValueError(),
                    err_msg="PYE configuration is required",
                )

            d = deploy_class(
                global_state.conf_data,
                global_state.cwd,
                Connection(**fabric_conf),
                global_state.env,
            )
            self.deploy_conn = d
            return d
        except FabikError as e:
            echo_error(e.err_msg)
            raise typer.Abort()
        except Exception as e:
            echo_error(str(e))
            raise typer.Abort()

    def __repr__(self) -> str:
        return f"""{self.__class__.__name__}(
    env={self.env}, 
    cwd={self.cwd!s}, 
    conf_file={self.conf_file!s}, 
    force={self.force}, 
    fabic_config={self.fabic_config!s}
)"""


# 创建全局状态实例并注册默认验证器
global_state = GlobalState()
global_state.register_config_validator(config_validator_name_workdir)


# ==============================================
# functions for command line interface
# ==============================================

NoteRename = Annotated[bool, typer.Option(help="If the target file exists, rename it.")]
NoteForce = Annotated[
    bool, typer.Option("--force", "-f", help="Force to overwrite the existing file.")
]
NoteEnvPostfix = Annotated[
    bool, typer.Option(help="在生成的配置文件名称末尾加上环境名称后缀。")
]
NoteReqirementsFileName = Annotated[
    str, typer.Option(help="指定 requirements.txt 的文件名。")
]


def main_callback(
    env: Annotated[
        str | None, typer.Option("--env", "-e", help="Environment name.")
    ] = None,
    cwd: Annotated[
        Path | None,
        typer.Option(
            file_okay=False, exists=True, help="Specify the local working directory."
        ),
    ] = None,
    conf_file: Annotated[
        Path | None,
        typer.Option(
            file_okay=True, exists=True, help="Specify the configuration file."
        ),
    ] = None,
):
    try:
        global_state.env = env
        global_state.cwd = FabikConfig.gen_work_dir(cwd)
        global_state.conf_file = (
            FabikConfig.gen_fabik_toml(work_dir=global_state.cwd, conf_file=None)
            if conf_file is None
            else conf_file
        )
    except FabikError as e:
        echo_error(e.err_msg)
        raise typer.Exit()

    echo_info(
        f"{global_state!r}",
        panel_title="LOG: main_callback",
    )


def main_init(
    full_format: Annotated[
        bool, typer.Option(help="Use full format configuration file.")
    ] = False,
    force: NoteForce = False,
):
    """[local] Initialize fabik project, create fabik.toml configuration file in the working directory."""
    work_dir: Path = global_state.check_work_dir_or_use_cwd()

    value = jinja2.Template(
        tpl.FABIK_TOML_TPL if full_format else tpl.FABIK_TOML_SIMPLE_TPL
    ).render(
        create_time=f"{fabik.__now__!s}",
        fabik_version=fabik.__version__,
        WORK_DIR=work_dir.absolute().as_posix(),
    )

    toml_file = work_dir.joinpath(tpl.FABIK_TOML_FILE)

    if toml_file.exists():
        if force:
            # 如果强制覆盖，直接写入文件
            echo_warning(f"{toml_file!s} already exists. Overwriting it.")
        else:
            # 如果不强制覆盖，提示文件存在并退出
            echo_warning(f"{toml_file!s} already exists. Use --force to overwrite it.")
            raise typer.Exit()

    # 写入配置文件
    toml_file.write_text(value)
    echo_info(f"{toml_file} has been created.")


def conf_callback(
    force: NoteForce = False,
):
    global_state.force = force


def conf_tpl(
    file: Annotated[
        list[str],
        typer.Argument(
            help="Provide configuration file names based on the tpl directory."
        ),
    ],
    env_postfix: NoteEnvPostfix = False,
):
    """[local] Initialize configuration file content based on the template files in the local tpl directory."""
    # 需要检查 tpl_dir 是否存在
    global_state.register_config_validator(config_validator_tpldir)
    global_state.load_conf_data(check=True)
    tpl_dir = Path(global_state.fabic_config.getcfg("PATH", "tpl_dir"))  # type: ignore
    tpl_names: list[str] = []
    for n in file:
        # 模板名称统一不带 jinja2 后缀，模板文件必须带有 jinja2 后缀。
        tpl_name = n[:-7] if n.endswith(".jinja2") else n
        tpl_file = tpl_dir.joinpath(f"{tpl_name}.jinja2")
        tpl_names.append(tpl_name)
        if not tpl_file.exists():
            echo_error(f'Template file "{tpl_file}" not found.')
            raise typer.Abort()

    for tpl_name in tpl_names:
        global_state.write_config_file(
            tpl_name,
            tpl_dir=tpl_dir,
            target_postfix=f".{global_state.env}" if env_postfix else "",
        )


def conf_make(
    file: Annotated[
        list[str],
        typer.Argument(help="Make a configuration file in fabik.toml."),
    ],
    env_postfix: NoteEnvPostfix = False,
):
    """[local] Initialize configuration file content based on the fabik.toml, no template file is used."""
    global_state.load_conf_data(check=True)
    for f in file:
        global_state.write_config_file(
            f, target_postfix=f".{global_state.env}" if env_postfix else ""
        )


# -------------------------------------- 子命令 gen
def gen_password(
    password: Annotated[str, typer.Argument(help="提供密码。", show_default=False)],
    salt: Annotated[
        str,
        typer.Argument(help="提供密码盐值。", show_default=False),
    ],
):
    """返回加盐之后的 PASSWORD。"""
    echo_info(util.gen.gen_password(password, salt))


def gen_fernet_key():
    """生成一个 SECRET_KEY。"""
    echo_info(util.gen.gen_fernet_key())


def gen_token(length: Annotated[int, typer.Argument(help="字符串位数。")] = 8):
    """根据提供的位数，返回一个 token 字符串。"""
    echo_info(util.gen.gen_token(k=length))


def gen_uuid(
    name: Annotated[UUIDType, typer.Argument(help="UUID 类型。")] = UUIDType.UUID4,
):
    """根据提供的名称，调用 uuid 库返回一个 UUID 字符串，仅支持 uuid1 和 uuid4 两种类型。"""
    echo_info(util.gen.gen_uuid(name.value))


def gen_requirements(
    force: NoteForce = False,
    requirements_file_name: NoteReqirementsFileName = "requirements.txt",
):
    """使用 uv 命令为当前项目生成 requirements.txt 依赖文件。"""
    work_dir: Path = global_state.check_work_dir_or_use_cwd()
    requirements_txt = work_dir / requirements_file_name
    if requirements_txt.exists() and not force:
        echo_warning(
            f"{requirements_txt.absolute().as_posix()} 文件已存在，使用 --force 强制覆盖。"
        )
        raise typer.Exit()
    try:
        # 执行 uv export 命令生成 requirements.txt
        subprocess.run(
            [
                "uv",
                "export",
                "--format",
                "requirements-txt",
                "--no-hashes",
                "--output-file",
                f"{requirements_txt.absolute().as_posix()}",
            ],
            capture_output=True,
            text=True,
            check=True,
        )
        echo_info(f"{requirements_txt.absolute().as_posix()} 文件已成功生成！")
    except subprocess.CalledProcessError as e:
        echo_error(f"生成 {requirements_txt.absolute().as_posix()} 失败: {e.stderr}")
        raise typer.Abort()
    except FileNotFoundError:
        echo_error("未找到 uv 命令，请确保已安装 uv 工具")
        raise typer.Abort()


# ---------------------------- 子命令 venv
#
def server_callback(
    deploy_class: Annotated[
        DeployClassName, typer.Option(help="指定部署类。")
    ] = DeployClassName.GUNICORN,
):
    if deploy_class == DeployClassName.GUNICORN:
        from fabik.deploy.gunicorn import GunicornDeploy as Deploy

        global_state.build_deploy_conn(Deploy)
    elif deploy_class == DeployClassName.uWSGI:
        from fabik.deploy.uwsgi import UwsgiDeploy as Deploy

        global_state.build_deploy_conn(Deploy)


def venv_init(
    requirements_file_name: NoteReqirementsFileName = "requirements.txt",
):
    """「远程」部署远程服务器的虚拟环境。"""
    try:
        rsync_exclude = global_state.conf_data.get("RSYNC_EXCLUDE", [])
        global_state.deploy_conn.rsync(exclude=rsync_exclude)
        global_state.deploy_conn.init_remote_venv(requirements_file_name)
    except Exception as e:
        echo_error(f"初始化虚拟环境失败: {str(e)}")
        raise typer.Abort()


def venv_update(
    name: Annotated[
        list[str] | None, typer.Argument(help="指定希望更新的 pip 包名称。")
    ] = None,
    all: Annotated[bool, typer.Option(help="更新所有 pip 包。")] = False,
):
    """「远程」部署远程服务器的虚拟环境。"""

    try:
        if all:
            global_state.deploy_conn.pipupgrade(all=True)
        elif name is not None and len(name) > 0:
            global_state.deploy_conn.pipupgrade(names=name)
        else:
            echo_error("请提供希望更新的 pip 包名称。")
            raise typer.Abort()
    except Exception as e:
        echo_error(f"更新 pip 包失败: {str(e)}")
        raise typer.Abort()


def venv_outdated():
    """「远程」打印所有的过期的 python package。"""
    global_state.deploy_conn.pipoutdated()  # type: ignore # noqa: F821


# ---------------------------- 子命令 server


def server_deploy():
    """「远程」部署项目到远程服务器。"""
    global_state.deploy_conn.rsync(
        exclude=global_state.fabic_config.getcfg("RSYNC_EXCLUDE", [])
    )  # type: ignore # noqa: F821
    global_state.deploy_conn.put_config(force=True)  # type: ignore # noqa: F821


def server_start():
    """「远程」在服务器上启动项目进程。"""
    global_state.deploy_conn.start()  # type: ignore # noqa: F821


def server_stop():
    """「远程」在服务器上停止项目进程。"""
    global_state.deploy_conn.stop()  # type: ignore # noqa: F821


def server_reload():
    """「远程」在服务器上重载项目进程。"""
    global_state.deploy_conn.reload()  # type: ignore # noqa: F821


def server_dar():
    """「远程」在服务器上部署代码，然后执行重载。也就是 deploy and reload 的组合。"""
    try:
        global_state.deploy_conn.rsync(
            exclude=global_state.conf_data.get("RSYNC_EXCLUDE", [])
        )  # type: ignore # noqa: F821
        global_state.deploy_conn.put_config(force=True)  # type: ignore # noqa: F821
        global_state.deploy_conn.reload()  # type: ignore # noqa: F821
    except FabikError as e:
        echo_error(e.err_msg)
        raise typer.Abort()
    except Exception as e:
        echo_error(str(e))
        raise typer.Abort()



"""
fabik is a library for Python command line application.
-----------------------
"""

__version__ = '25.7.1'
__author__ = 'zrong'
__name__ = 'fabik'
__email__ = 'zrongzrong@gmail.com'
__url__ = 'https://github.com/zrong/fabik'
__license__ = 'MIT'
__description__ = 'fabik is a library for Python command line application.'

import datetime
__now__ = datetime.datetime.now()

__all__ = [
    '__version__',
    '__author__',
    '__email__',
    '__url__',
    '__license__',
    '__description__',
    '__now__',
]




"""  .._fabik_cli:

fabik.cli
----------------------------
    
fabik command line interface implementation
"""

import typer
from fabik.cmd import (
    main_callback,
    main_init,
    conf_callback,
    conf_tpl,
    conf_make,
    gen_password,
    gen_fernet_key,
    gen_token,
    gen_uuid,
    gen_requirements,
    server_callback,
    venv_init,
    venv_update,
    venv_outdated,
    server_deploy,
    server_start,
    server_stop,
    server_reload,
    server_dar,
)

main: typer.Typer = typer.Typer()

sub_gen: typer.Typer = typer.Typer(name='gen', help='[local] Generate common strings.')
sub_conf: typer.Typer = typer.Typer(name='conf', help='[local/remote] Process configuration files.')
sub_venv: typer.Typer = typer.Typer(
    name='venv', help='[remote] Process Python virtual environment on remote server.'
)
sub_server: typer.Typer = typer.Typer(
    name='server', help='[remote] Process remote server.'
)

main.add_typer(sub_gen)
main.add_typer(sub_conf)
main.add_typer(sub_venv)
main.add_typer(sub_server)


# The method in cmd module may be used as a command by other modules, so it is not decorated with a decorator.
main.callback()(main_callback)
main.command('init')(main_init)

sub_conf.callback()(conf_callback)
sub_conf.command('tpl')(conf_tpl)
sub_conf.command('make')(conf_make)

sub_gen.command('password')(gen_password)
sub_gen.command('fernet-key')(gen_fernet_key)
sub_gen.command('token')(gen_token)
sub_gen.command('uuid')(gen_uuid)
sub_gen.command('requirements')(gen_requirements)
    
    
sub_venv.callback()(server_callback)
sub_venv.command('init')(venv_init)
sub_venv.command('update')(venv_update)
sub_venv.command('outdated')(venv_outdated)


sub_server.callback()(server_callback)
sub_server.command('deploy')(server_deploy)
sub_server.command('start')(server_start)
sub_server.command('stop')(server_stop)
sub_server.command('reload')(server_reload)
sub_server.command('dar')(server_dar)


"""
.. _fabik_tpl:

fabik.tpl
-------------------

Provide template support.
"""


FABIK_TOML_FILE: str = 'fabik.toml'
""" Main config file name. """


FABIK_TOML_SIMPLE_TPL: str = """
###########################################
# fabik main config file
#
# @author:  zrong(https://zengrong.net)
# @source:  https://github.com/zrong/fabik.git
# @create:  {{ create_time }}
# @version: {{ fabik_version }}
# 
###########################################
# use NAME to define your project name.
NAME = '{{ NAME }}'

# Define all PATH related variables here.
[PATH]
# Define project source code folder, default is the absolute path of fabik command run folder.
work_dir = "{{ WORK_DIR }}"

# Define template folder, default is the absolute path of fabik command run folder + '/tpls'.
tpl_dir = "{{ WORK_DIR }}/tpls"
"""
""" This is a template for fabik main config file. """


FABIK_TOML_TPL: str = """
###########################################
# fabik main config file
#
# @author:  zrong(https://zengrong.net)
# @source:  https://github.com/zrong/fabik.git
# @create:  {{ create_time }}
# @version: {{ fabik_version }}
# 
#
# Use { { } } to include variable names, which will be replaced with specific values, two curly braces should be continuous, no content should be included, such as no spaces.
#
# NAME: project name, defined in config file
# WORK_DIR: absolute path of project source code folder
# DEPLOY_DIR: absolute path of deploy folder on server
###########################################

# use NAME to define your project name.
NAME = '{{ NAME }}'

# Python executable name, default is 'python3', you can use a absolute path to specify a specific python executable.
PYE = 'python3'

# DEPLOY_DIR is the absolute path of deploy folder on server
DEPLOY_DIR = '/srv/app/{{ NAME }}'

# Allow to replace environment variables.
# The following { { } } variables will be replaced with environment variables. Two curly braces should be continuous, no content should be included, such as no spaces.
# If the value contains the following keys, it will be replaced with the value of the environment variable.
#
# For example:
# 1. The project name is 'fabik' (default value), when it is replaced with environment variable, it will be converted to uppercase.
# 2. The environment variable contains FABIK_LOCAL_ADMIN_NAME
# 3. When using --env local to generate the config file, it will replace the value of {{ ADMIN_NAME }} with the value of the environment variable FABIK_LOCAL_ADMIN_NAME
#
# ADMIN_NAME
# ADMIN_PASSWORD
# SECRET_KEY
#
# You can add environment variables yourself, as long as the variable name in the config file is the same.
REPLACE_ENVIRON = [
    'ADMIN_NAME',
    'ADMIN_PASSWORD',
]

# Define all PATH related variables here.
[PATH]
# Define project source code folder, default is the absolute path of fabik command run folder.
work_dir = "{{ WORK_DIR }}"

# Define template folder, default is the absolute path of fabik command run folder + '/tpls'.
tpl_dir = "{{ WORK_DIR }}/tpls"


#==============================================
# The following configuration overrides the above configuration.
# The override is performed in the form of "supplement", the same name key will be overwritten, and the different name key will be retained.
#
# ENV represents the environment, when you want to deploy to different environments, you can use these different environments
# to generate configuration files for different environments, the key name after ENV is the environment name.
#
# For example, [ENV.local.path] is used to replace the path.work_dir or path.tpl_dir to specific value in the local environment.
#==============================================
[ENV.local.PATH]
work_dir = "/Users/zrong/tool/my_new_tool"
tpl_dir = "/Users/zrong/tool/my_new_tool/tpls"

"""
""" This is a template for fabik main config file. """



""" .._fabik_util_gen:

fabik.util.gen
~~~~~~~~~~~~~~~~~~~

生成器
"""

import string
import uuid
import random
from datetime import datetime

from fabik.util import md5txt
from fabik.util.encrypt import Encrypt


def gen_token(k=8, chars=string.ascii_letters + string.digits):
    """ 生成一段加密用的字符串
    字符串可以被用于密码、兑换码等等
    """
    return ''.join(random.choices(chars, k=k))


def gen_password(password: str, salt: str):
    """
    计算密码散列后的值
    :param password:
    :param salt:
    :return:
    """
    return md5txt(md5txt(str(password)) + str(salt))


def gen_random_name(value=None, prefix=''):
    """ 提供一个整数和一个前缀，获取一个随机的昵称 """
    if not isinstance(value, int):
        ts = str(datetime.now().timestamp()).split('.')
        value = int(ts[0]) + int(ts[1])
    # 将 value 转成16进制然后倒序
    return prefix + hex(value)[2:][::-1]


# 允许用于 REDIS 的数据类型
DATA_TYPES = ['list', 'string', 'zset', 'hash']
def gen_redis_key_name(r, name, prefix=None, postfix=None, data_type=None, extra=[]):
    """ 生成一个可用的名称，这个名称可能用于配置的键名，或者是排行榜的键名
    
    :param r: regional
    :param name: 名称
    :param prefix: 前缀
    :param postfix: 后缀
    :param date_type: 数据类型。数据类型位于后缀之后
    :param extra: 附加数据
    """
    names = []
    if prefix is not None:
        names.append(str(prefix))
    names.append(str(r))
    names.append(str(name))
    if postfix is not None:
        names.append(str(postfix))
    if data_type in DATA_TYPES:
        names.append(data_type)
    return ':'.join(names + extra)


def gen_sign(**kwargs):
    """ 根据传递来的参数顺序生成校验码
    """
    keys = sorted(kwargs.keys(), key=str.lower)
    k_eq_v = []
    for k in keys:
        value = kwargs[k]
        # 复杂对象、空值、布尔值不参加校验
        if isinstance(value, list) or \
            isinstance(value, dict) or \
            isinstance(value, bool) or \
            value is None:
            continue
        k_eq_v.append(str(k) + '=' + str(value))
    return md5txt('&'.join(k_eq_v))


def gen_fernet_key() -> str:
    """ 使用 cryptography.fernet 生成一个 Fernet 算法 KEY。
    """
    return Encrypt.fernet_key().decode()


def gen_uuid(name: str= 'uuid4') -> str:
    """ 使用 uuid 库生成一个 UUID。
    """
    if name == 'uuid1':
        return str(uuid.uuid1())
    return str(uuid.uuid4())
        


""" .._fabik_util_jinja_filter:

fabik.util.jinja_filter
~~~~~~~~~~~~~~~~~~~

自定义 jinja filter

@author zrong
@created 2023-11-22
"""

from typing import Any
from datetime import datetime


def _jinja_filter_check_int_str(value: Any) -> Any:
    """ 检测 value 如果是 None 或者空字符串，就返回空字符串。"""
    if value is None:
        return None
    if isinstance(value, str) and len(value.strip()) == 0:
        return None
    return value


def jinja_filter_strftimestamp(ts: float | int | str | None, fmt: str | None = None):
    """将 timestamp 转换成为字符串。"""
    # fmt = '%Y-%m-%d'
    ts = _jinja_filter_check_int_str(ts)
    if ts is None:
        return ''
    ts = float(ts)
    if ts <= 0:
        return ''
    dt = datetime.fromtimestamp(ts)
    if fmt is None:
        return dt.isoformat()
    return dt.strftime(fmt)


_KB: int = 1024
_MB: int = 1048576
_GB: int = 1073741824
def jinja_filter_filesize(size: int | str | None) -> Any:
    """ 将文件大小（字节）转换为人类可读的形式。"""
    size = _jinja_filter_check_int_str(size)
    if size is None:
        return ''
    size = int(size)
    if size > _GB:
        return f'{size/_GB:.2f} GB'
    elif size > _MB:
        return f'{size/_MB:.2f} MB'
    elif size > _KB:
        return f'{size/_KB:.1f} KB'
    return f'{size} Byte'



""" .. _fabik_deploy:

fabik.deploy
~~~~~~~~~~~~~~~~~~~

封装 fabric 的功能，提供远程部署能力。
"""

import re
import sys
import logging
import json
from datetime import datetime
from pathlib import Path

from invoke import runners
from fabric.connection import Connection
from invoke.exceptions import Exit

from fabik.conf import ConfigReplacer
from fabik.error import FabikError, echo_error

logger = logging.Logger('fabric', level=logging.DEBUG)
logger.addHandler(logging.StreamHandler(sys.stdout))


def rsync(
    c,
    source: str,
    target: str,
    exclude=(),
    delete: bool=False,
    strict_host_keys: bool=True,
    rsync_opts: str="",
    ssh_opts: str="",
):
    """
    Convenient wrapper around your friendly local ``rsync``.

    https://github.com/fabric/patchwork/blob/master/patchwork/transfers.py

    Specifically, it calls your local ``rsync`` program via a subprocess, and
    fills in its arguments with Fabric's current target host/user/port. It
    provides Python level keyword arguments for some common rsync options, and
    allows you to specify custom options via a string if required (see below.)

    For details on how ``rsync`` works, please see its manpage. ``rsync`` must
    be installed on both the invoking system and the target in order for this
    function to work correctly.

    .. note::
        This function transparently honors the given
        `~fabric.connection.Connection`'s connection parameters such as port
        number and SSH key path.

    .. note::
        For reference, the approximate ``rsync`` command-line call that is
        constructed by this function is the following::

            rsync [--delete] [--exclude exclude[0][, --exclude[1][, ...]]] \\
                -pthrvz [rsync_opts] <source> <host_string>:<target>

    :param c:
        `~fabric.connection.Connection` object upon which to operate.
    :param str source:
        The local path to copy from. Actually a string passed verbatim to
        ``rsync``, and thus may be a single directory (``"my_directory"``) or
        multiple directories (``"dir1 dir2"``). See the ``rsync`` documentation
        for details.
    :param str target:
        The path to sync with on the remote end. Due to how ``rsync`` is
        implemented, the exact behavior depends on the value of ``source``:

        - If ``source`` ends with a trailing slash, the files will be dropped
          inside of ``target``. E.g. ``rsync(c, "foldername/",
          "/home/username/project")`` will drop the contents of ``foldername``
          inside of ``/home/username/project``.
        - If ``source`` does **not** end with a trailing slash, ``target`` is
          effectively the "parent" directory, and a new directory named after
          ``source`` will be created inside of it. So ``rsync(c, "foldername",
          "/home/username")`` would create a new directory
          ``/home/username/foldername`` (if needed) and place the files there.

    :param exclude:
        Optional, may be a single string or an iterable of strings, and is
        used to pass one or more ``--exclude`` options to ``rsync``.
    :param bool delete:
        A boolean controlling whether ``rsync``'s ``--delete`` option is used.
        If True, instructs ``rsync`` to remove remote files that no longer
        exist locally. Defaults to False.
    :param bool strict_host_keys:
        Boolean determining whether to enable/disable the SSH-level option
        ``StrictHostKeyChecking`` (useful for frequently-changing hosts such as
        virtual machines or cloud instances.) Defaults to True.
    :param str rsync_opts:
        An optional, arbitrary string which you may use to pass custom
        arguments or options to ``rsync``.
    :param str ssh_opts:
        Like ``rsync_opts`` but specifically for the SSH options string
        (rsync's ``--rsh`` flag.)
    """
    # Turn single-string exclude into a one-item list for consistency
    if isinstance(exclude, str):
        exclude = [exclude]
    # Create --exclude options from exclude list
    exclude_opts = ' --exclude "{}"' * len(exclude)
    # Double-backslash-escape
    exclusions = tuple([str(s).replace('"', '\\\\"') for s in exclude])
    # Honor SSH key(s)
    key_string = ""
    # TODO: seems plausible we need to look in multiple places if there's too
    # much deferred evaluation going on in how we eg source SSH config files
    # and so forth, re: connect_kwargs
    # TODO: we could get VERY fancy here by eg generating a tempfile from any
    # in-memory-only keys...but that's also arguably a security risk, so...
    keys = c.connect_kwargs.get("key_filename", [])
    # TODO: would definitely be nice for Connection/FabricConfig to expose an
    # always-a-list, always-up-to-date-from-all-sources attribute to save us
    # from having to do this sort of thing. (may want to wait for Paramiko auth
    # overhaul tho!)
    if isinstance(keys, str):
        keys = [keys]
    if keys:
        key_string = "-i " + " -i ".join(keys)
    # Get base cxn params
    user, host, port = c.user, c.host, c.port
    port_string = f"-p {port}"
    # Remote shell (SSH) options
    rsh_string = ""
    # Strict host key checking
    disable_keys = "-o StrictHostKeyChecking=no"
    if not strict_host_keys and disable_keys not in ssh_opts:
        ssh_opts += f" {disable_keys}"
    rsh_parts = [key_string, port_string, ssh_opts]
    if any(rsh_parts):
        rsh_string = "--rsh='ssh {}'".format(" ".join(rsh_parts))
    # Set up options part of string
    options_map = {
        "delete": "--delete" if delete else "",
        "exclude": exclude_opts.format(*exclusions),
        "rsh": rsh_string,
        "extra": rsync_opts,
    }
    options = "{delete}{exclude} -pthrvz {extra} {rsh}".format(**options_map)
    # Create and run final command string
    # TODO: richer host object exposing stuff like .address_is_ipv6 or whatever
    if host.count(":") > 1:
        # Square brackets are mandatory for IPv6 rsync address,
        # even if port number is not specified
        cmd = "rsync {} {} [{}@{}]:{}"
    else:
        cmd = "rsync {} {} {}@{}:{}"
    cmd = cmd.format(options, source, user, host, target)
    return c.local(cmd)


class Deploy(object):
    fabik_conf: dict
    work_dir: Path
    conn: Connection
    env_name: str | None = None
    pye: str
    replacer: ConfigReplacer

    def __init__(self, fabik_conf: dict, work_dir: Path, conn: Connection, env_name: str | None=None):
        """ 初始化
        """
        self.env_name = env_name
        self.fabik_conf = fabik_conf
        self.conn = conn
        self.work_dir = Path(work_dir)
        self.pye = fabik_conf['PYE']
        self.replacer = ConfigReplacer(fabik_conf, self.work_dir, env_name=env_name)
        self.replacer.check_env_name()

    def check_remote_conn(self):
        """ 确保当前提供的 conn 是远程版本
        """
        if not isinstance(self.conn, Connection):
            raise Exit('Use -H to provide a host!')

    def get_remote_path(self, *args) -> str:
        return self.replacer.deploy_dir.joinpath(*args).as_posix()

    def remote_exists(self, file):
        """ 是否存在远程文件 file
        """
        self.check_remote_conn()
        # files.exists 仅接受字符串
        if isinstance(file, Path):
            file = file.resolve().as_posix()
        command = f'test -e "$(echo {file})"'
        logger.info(f'{command=}')
        return self.conn.run(command, hide=True, warn=True).ok

    def make_remote_dir(self, *args):
        """ 创建部署文件夹
        """
        self.check_remote_conn()
        remotedir = self.get_remote_path(*args)
        if not self.remote_exists(remotedir):
            command = 'mkdir %s' % remotedir
            logger.info('创建远程文件夹 %s', command)
            self.conn.run(command)

    def cat_remote_file(self, *args):
        """ 使用 cat 命令获取远程文件的内容
        """
        f = self.get_remote_path(*args)
        logger.info('cat_remote_file %s', f)
        if not self.remote_exists(f):
            return None
        result = self.conn.run('cat ' + f, warn=False, hide=True)
        return result.stdout

    def get_remote_pid(self, host=None, port=None):
        """ 利用命令行查找某个端口运行进程的 pid
        :param host: IP 地址
        :param port: 端口号
        """
        self.check_remote_conn()
        address = None
        if host:
            address = host
        if port:
            p = ':' + str(port)
            if address:
                address += '@' + p
            else:
                address = p
        if not address:
            raise Exit('需要 host 或 port 配置。')
        # command = 'lsof -i :2004 | tail -1'
        command_fmt = 'lsof -i {} | tail -1'
        command = command_fmt.format(address)
        result = self.conn.run(command, warn=False, hide=True)
        if result.stdout == '':
            return None
        return re.split(r'\s+', result.stdout)[1]

    def init_remote_dir(self, deploy_dir):
        """ 创建远程服务器的运行环境
        """
        deploy_dir_path = Path(deploy_dir)
        for d in [ self.replacer.deploy_dir,
            deploy_dir,
            deploy_dir_path.joinpath('logs'),
            deploy_dir_path.joinpath('output') ]:
            self.make_remote_dir(d)

    def source_venv(self):
        remote_venv_dir = self.get_remote_path('venv')
        if not self.remote_exists(remote_venv_dir):
            raise Exit('venv 还没有创建！请先执行 init_remote_venv')
        return f'source {remote_venv_dir}/bin/activate'

    def init_remote_venv(self, requirements_file_name: str):
        """ 创建虚拟环境
        """
        remote_venv_dir = self.get_remote_path('venv')
        
        # 检查 Python 是否可用
        python_check = self.conn.run(f'{self.pye} --version', hide=True, warn=True)
        if not python_check.ok:
            raise Exit(f'Python 可执行文件 {self.pye} 未找到或未安装')
        
        # 创建虚拟环境（如果不存在）
        if not self.remote_exists(remote_venv_dir):
            venv_result = self.conn.run(f'{self.pye} -m venv {remote_venv_dir}', warn=True)
            if not venv_result.ok:
                # 尝试使用 python3 或 python
                alt_python = 'python3' if self.pye != 'python3' else 'python'
                venv_result = self.conn.run(f'{alt_python} -m venv {remote_venv_dir}', warn=True)
                if not venv_result.ok:
                    raise Exit(f'创建虚拟环境失败: {venv_result.stderr}')
        
        # 检查虚拟环境是否成功创建
        if not self.remote_exists(f'{remote_venv_dir}/bin/activate'):
            raise Exit('虚拟环境创建失败，激活脚本不存在')
            
        # 确保 pip 可用并更新
        with self.conn.prefix(f'source {remote_venv_dir}/bin/activate'):
            # 先确保 pip 存在
            pip_check = self.conn.run('pip --version', hide=True, warn=True)
            if not pip_check.ok:
                # 尝试安装 pip
                self.conn.run('curl https://bootstrap.pypa.io/get-pip.py | python', warn=True)
            
            # 更新 pip
            pip_update = self.conn.run('pip install -U pip', warn=True)
            if not pip_update.ok:
                logger.warning('更新 pip 失败，继续执行后续步骤')
            
            # 安装 requirements（如果文件存在）
            req_file = self.get_remote_path(requirements_file_name)
            if self.remote_exists(req_file):
                self.conn.run(f'pip install -r {req_file}', warn=True)
            else:
                logger.warning(f'未找到 requirements 文件: {req_file}')

    def piplist(self, format='columns'):
        """ 获取虚拟环境中的所有安装的 python 模块
        :@param format: columns (default), freeze, or json
        """
        with self.conn.prefix(self.source_venv()):
            result = self.conn.run('pip list --format ' + format)
            return result.stdout

    def pipoutdated(self, format='columns'):
        """ 查看过期的 python 模块
        :@param format: columns (default), freeze, or json
        """
        with self.conn.prefix(self.source_venv()):
            result = self.conn.run('pip list --outdated --format ' + format)
            return result.stdout

    def pipupgrade(self, names=None, all=False):
        """ 更新一个 python 模块
        """
        with self.conn.prefix(self.source_venv()):
            mod_names = []
            if all:
                result = self.conn.run('pip list --outdated --format json')
                if result.ok:
                    mod_names = [item['name'] for item in json.loads(result.stdout)]
            elif names:
                mod_names = [name for name in names]
            if mod_names:
                self.conn.run('pip install -U ' + ' '.join(mod_names))

    def put_tpl(self, tpl_name, force=False):
        """ 基于 jinja2 模板生成配置文件，根据 env 的值决定是否上传
        """
        # 创建远程文件夹
        self.make_remote_dir()
        # 获取远程文件的绝对路径
        target_remote = self.get_remote_path(tpl_name)
        tpltarget_remote_exists = self.remote_exists(target_remote)
        if force and tpltarget_remote_exists:
            logger.warning('delete %s', target_remote)
            remoter = self.conn.run(f'rm -f {target_remote}')
            if remoter.ok:
                logger.warning(f'删除远程配置文件 {target_remote}')
            tpltarget_remote_exists = False
        
        # 本地创建临时文件后上传
        if force or not tpltarget_remote_exists:
            # 创建一个临时文件用于上传，使用后缀
            _, final_file = self.replacer.set_writer(tpl_name, force=force, target_postfix=f'.{self.env_name}')
            self.conn.put(final_file, target_remote)
            logger.warning('覆盖远程配置文件 %s', target_remote)
            localrunner = runners.Local(self.conn)
            # 删除本地的临时配置文件
            localr = localrunner.run(f'rm -f {final_file.as_posix()}')
            if localr is not None and localr.ok:
                logger.warning(f'删除本地临时文件 {final_file.as_posix()}')
            
    def put_config(self, files: dict[str, str] | None = None, force: bool=False) -> None:
        """上传配置文件到远程服务器
        
        :param files: 要上传的配置文件字典，如果为None则使用默认配置文件
        :param force: 是否强制覆盖已存在的文件
        """
        if files is None:
            # 默认上传常用配置文件
            default_files = {
                '.env': '.env',
                'gunicorn.conf.py': 'gunicorn.conf.py',
                'config.toml': 'config.toml'
            }
            files = default_files
            
        for tpl_name in files.keys():
            self.put_tpl(tpl_name, force)

    def rsync(self, exclude=[], is_windows=False):
        """ 部署最新程序到远程服务器
        """
        if is_windows:
            # 因为 windows 下面的 rsync 不支持 windows 风格的绝对路径，转换成相对路径
            pdir = str(self.work_dir.relative_to('.').resolve())
        else:
            pdir = str(self.work_dir.resolve())
        if not pdir.endswith('/'):
            pdir += '/'
        deploy_dir = self.get_remote_path()
        self.init_remote_dir(deploy_dir)
        rsync(self.conn, pdir, deploy_dir, exclude=exclude)
        logger.warn('RSYNC [%s] to [%s]', pdir, deploy_dir)

    def get_logs(self, extras=[]):
        """ 下载远程 logs 到本地
        """
        log_files = ['app.log', 'error.log', 'access.log']
        time_string = datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        for f in log_files + extras:
            logf = self.get_remote_path('logs/{}'.format(f))
            if not self.remote_exists(logf):
                logger.warning('找不到远程 log 文件 %s', logf)
                continue
            logp = Path(logf)
            local_file = self.work_dir.joinpath('logs', '{name}_{basename}_{times}{extname}'.format(name=self.env_name, times=time_string, basename=logp.name, extname=logp.suffix))
            self.conn.get(logf, local=local_file)


""" .._fabik_util_func:

fabik.util.func
~~~~~~~~~~~~~~~~~~~

工具类，可独立使用。
放置和请求完全无关的，不依赖任何 flask 框架内容的工具。
这个类中的方法可以被其它模块导入而不发生冲突。
"""

from typing import Union
from datetime import datetime, timedelta, date, time as time2
import re
import json


_simple_class = [int, float, str, bool, dict, list, datetime]


def issimple(value):
    """ 判断一个值是否为简单类型，包括 None
    """
    for typeclass in _simple_class:
        if value is None or isinstance(value, typeclass):
            return True
    return False


def parse_bool(value, default_value=False):
    if value is not None:
        return value.lower() == 'true'
    return default_value


def parse_int(value, default_value=None):
    if value is not None:
        try:
            value = int(value)
            return value
        except:
            return default_value
    return default_value


def parse_float(value, default_value=None):
    if value is not None:
        try:
            value = float(value)
            return value
        except:
            return default_value
    return default_value


def parse_date(value, default_value=None, to_day_end=False) -> datetime:
    """ 将一个日期字符串解析成为 datetime 对象。

    :param value:
    :param default_value:
    :param to_day_end:  值为True 则自动加上23小时59分59秒999999微秒（当天的最后时刻），用来判断包含关系。
    :return:
    """
    if value is None:
        value = default_value
    else:
        try:
            value = datetime.strptime(value, '%Y-%m-%d')
        except:
            value = default_value
    if isinstance(value, datetime) and to_day_end:
        # 86399 = 3600*24-1
        value += timedelta(seconds=86399, microseconds=999999)
    return value


def between_date(from_date, to_date) -> list[date]:
    """ 传递一个 from_date 和 一个 to_date，返回中间所有的 date（包括 from_date 和 to_date）。
    """
    fd = date(from_date.year, from_date.month, from_date.day)
    td = date(to_date.year, to_date.month, to_date.day)
    days = [fd]
    # from 和 to 是同一天
    if fd == td:
        return days
    # from 在 to 之后，以 from 为准
    if fd > td:
        return days
    ld = date(fd.year, fd.month, fd.day)
    while ld < td:
        ld += timedelta(days=1)
        days.append(ld)
    return days


def strfdate(dtobj):
    """ 将一个 datetime 转换成成 YYYY-MM-DD 形式的字符串
    :param dtobj: datetime 实例
    :return:
    """
    return dtobj.strftime('%Y-%m-%d')


def strptime(timestr):
    """ 将一个 time 字符串按照 HH:MM:SS 形式的字符串转换成 datetime.time 对象
    :param timestr: 时间字符串
    :return: datetime.time
    """
    dt = datetime.strptime(timestr, '%H:%M:%S')
    return time2(dt.hour, dt.minute, dt.second)


def in_time_range(timestart: Union[str, time2], timeend: Union[str, time2], dt: datetime=None) -> bool:
    """ 判断给定的时间 dt 是否在两个时间中间。

    :param timestart: 开始的时间，支持字符串和 time 对象，字符串使用本模块中的 strptime 解析。
    :param timeend: 结束的时间，支持字符串和 time 对象，字符串使用本模块中的 strptime 解析。
    :param dt: 给定的 datetime 对象，如果不提供则使用 ``datetime.now()``。
    """
    if isinstance(timestart, str):
        timestart = strptime(timestart)
    if isinstance(timeend, str):
        timeend = strptime(timeend)
    if dt is None:
        dt = datetime.now()
    dtstart = datetime(dt.year, dt.month, dt.day, timestart.hour, timestart.minute, timestart.second)
    dtend = datetime(dt.year, dt.month, dt.day, timeend.hour, timeend.minute, timeend.second)
    # 如果 timeend 小于 timestart，就将 timeend 增加一天。
    if timeend < timestart:
        dtend += timedelta(days=1)
    return dt > dtstart and dt < dtend


def next_month_dt(day: datetime=None) -> datetime:
    """ 返回下一个月的 datetime 对象。

    :param day: 基准 datetime 对象
    """
    if day is None:
        day = date.today()
    m = day.month + 1
    y = day.year
    if m > 12:
        # 处理跨年
        m = 1
        y += 1
    return datetime(y, m, 1)


def last_month_dt(day: datetime=None) -> datetime:
    """ 返回上个月的 datetime 对象。

    :param day: 基准 datetime 对象。
    """
    if day is None:
        day = date.today()
    # 找到上个月
    m = day.month - 1
    y = day.year
    if m <= 0:
        # 处理跨年
        m = 12
        y -= 1
    return datetime(y, m, 1)


def daydt(day: Union[date, datetime]=None,
    default_initday: bool=False,
    first_day_of_month: bool=False) -> datetime:
    """ 获取一个hour为0的日期。

    :param day: 时间戳或者基准 date/datetime 对象。
    :param default_initday: 仅当 day 参数为 None 的时候有效，True 代表使用初始日期，否则使用今天。
    :param first_day_of_month: 值为 True 则返回提供的 day 所在月的第一天。
    :return:
    """
    if not day:
        if default_initday:
            day = date(2017, 10, 15)
        else:
            day = date.today()
    elif isinstance(day, int):
        day = datetime.fromtimestamp(day)
    return datetime(day.year, day.month, 1 if first_day_of_month else day.day)


def get_cur_and_next_month(day: datetime=None):
    """ 获取代表本月到下月的日期对象。

    :param day: 一个 datetime 对象，基于这个对象作为基准来创建 cur_month。
    """
    # 如果不传递则使用今天作为基准
    if day is None:
        day = date.today()
    # 今天所在的月的第一天
    cur_month = datetime(day.year, day.month, 1)
    return cur_month, next_month_dt(cur_month)


def _json_datetime_handler(x):
    if isinstance(x, datetime):
        return x.isoformat()
    raise TypeError("Unknown type")


def jsondumps(obj) -> str:
    """ dump 一个 json 字符串，处理 datetime。

    :param obj: 需要被 dump 的对象。
    """
    return json.dumps(obj, default=_json_datetime_handler)


PVER_RE = re.compile(r'([abr])\d+', re.I)


def vername2code(vername) -> int:
    """ 转换 vername 到 vercode，每个版本位允许3位。

    :param vername: 形如 ``3.1.455`` 的版本号。
    """
    try:
        verlist = vername.split('.')
        vercode = 0
        if len(verlist) > 2:
            vercode += int(verlist[2])
        if len(verlist) > 1:
            vercode += int(verlist[1]) * 1000
        vercode += int(verlist[0]) * 1000000
        return vercode
    except Exception as e:
        return -1




""" .. _fabik_http:

fabik.http
----------------------

封装 httpx，自动处理响应。
"""
import json
from typing import Any
from collections.abc import Callable
import httpx
import asyncio


class ResponseValue:
    """ 封装 HTTP 响应数据。"""
    error: bool = False
    code: int = 200
    message: str = None

    value: dict = None
    httpx_response: httpx.Response = None
    request_value = None

    def __init__(self,
                 message: str = None,
                 *,
                 error: bool = None,
                 code: int = None,
                 value: dict = None,
                 httpx_response: httpx.Response = None,
                 request_value=None,
                 **kwargs) -> None:
        self.message = message
        # 不提供 message，也不提供 error，就是无错误 error=False, code=200
        if self.message is None:
            self.error = False if error is None else error
            self.code = 200 if code is None else code
        else:
            self.error = error
            self.code = code
        if self.error is None:
            self.error = True
        if self.code is None:
            self.code = 444

        self.value = value or {}
        self.httpx_response = httpx_response
        self.request_value = request_value
        if kwargs:
            self.value.update(kwargs)

    @property
    def text(self) -> str:
        if self.httpx_response:
            return self.httpx_response.text
        return None

    @property
    def content(self) -> bytes:
        if self.httpx_response:
            return self.httpx_response.content
        return None

    @property
    def headers(self) -> bytes:
        if self.httpx_response:
            return self.httpx_response.headers
        return None

    def get_value(self, key: str) -> Any:
        """ 获取 values 对象中保存的数据。"""
        if not self.value:
            return None
        return self.value.get(key)

    def to_dict(self,
                plain_value: bool = False,
                value_keyname: str = 'value',
                include_request_value: bool = False) -> dict:
        """ 将 ResponeValue 的数据作为 dict 导出。

        :param plain_value: 将 value 的值拉平。
        :param value_keyname: 指示 ResponseValue 中的 value 对象，在 dict 中的键名。
            若 plain_value 为 True，则这个值无效。
        """
        d = {
            'error': self.error,
            'code': self.code,
        }
        if self.message:
            d['message'] = self.message
        if self.value:
            if plain_value:
                d.update(self.value)
            else:
                d[value_keyname] = self.value
        if self.request_value and include_request_value:
            d['request_value'] = self.request_value.request_args
        return d

    def parse_and_merge_json(self,
                             *,
                             data_keyname: str = 'data',
                             code_keyname: str = 'code',
                             message_keyname: str = 'msg',
                             success_code_values: tuple[Any] = (0, 200, None),
                             value_keyname: str = 'value',
                             plain_value: bool = False) -> dict:
        """ 解析 http 响应中的 json 返回，将其中的有效数据合并到 ResponseValue 中。

        json 中一般会封装一个键名用于表现数据。以抖音 API 的返回为例：
        https://developer.open-douyin.com/docs/resource/zh-CN/mini-game/develop/server/interface-request-credential/get-access-token/
        >>> {
        >>>    "err_no": 40017,
        >>>    "err_tips": "bad secret",
        >>>    "data": {
        >>>             "access_token": "",
        >>>             "expires_in": 0
        >>>    }
        >>> }
        在上例中，data_keyname=data, code_keyname=err_no, message_keyname=err_tips。

        根据微信 API 文档，success_code_values 应该包含 None 这个值。
        > 注意：当API调用成功时，部分接口不会返回 errcode 和 errmsg，只有调用失败时才会返回。
        https://developers.weixin.qq.com/minigame/dev/guide/base-ability/backend-api.html

        :param data_keyname: api 返回的数据键名。
        :param code_keyname: api 返回的错误码键名。
        :param message_keyname: api 返回的错误消息键名。
        :param success_code_values: 哪些错误码代表调用成功。默认值为 (0, 200, None)。
        :param value_keyname: 同 to_dict() 定义。
        :param plain_value: 同 to_dict() 定义。
        """
        try:
            parsed_json = self.httpx_response.json()
            if parsed_json.get(code_keyname) is not None:
                self.code = parsed_json.get(code_keyname)
            # 代表上一步请求的 http 调用错误，也可能是解析后的 json 中有错误码
            if not self.code in success_code_values:
                self.error = True
                if parsed_json.get(message_keyname):
                    self.message = parsed_json.get(message_keyname)
            if data_keyname is None or parsed_json.get(data_keyname) is None:
                self.value = parsed_json
            else:
                self.value = parsed_json.get(data_keyname)
            return self.to_dict(plain_value=plain_value,
                                value_keyname=value_keyname)
        except Exception as e:
            self.error = True
            self.code = 500
            self.message = f'{e=}, {self.httpx_response.text=}' if self.httpx_response else str(
                e)

        return self.to_dict(plain_value=True)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} {self.to_dict()}'


class RequestValue:
    """ 封装 HTTP 请求数据。"""

    method: str = None
    """ 请求方法。"""

    url: str = None
    """ 请求路径。"""

    url_map: dict = None
    """ 若存在，执行 url = url.format_map(url_map) 。"""

    value: dict = None
    """ 请求值，在 GET 的时候作为 params，在 post 的时候作为 data。"""

    headers: dict = None
    """ 请求头。"""

    is_json: bool = False
    """ 代表这次使用 json 请求。"""

    check_response: Callable[[httpx.Response], ResponseValue] = None
    """ check_response 接收一个 httpx.Response 参数，返回 ResponseValue，用于下一步处理。"""

    after_response: Callable = None
    """ after_response 保存期待在获得 ResponseValue 之后再调用的方法。
    check_response 可以修改原始的 httpx.Respone，
    而 after_response 更加灵活，对参数和返回值都不做限制。
    """

    kwargs: dict = None

    def __init__(self,
                 *,
                 url: str,
                 method: str = 'GET',
                 value: dict = None,
                 url_map: dict = None,
                 headers: dict = None,
                 is_json: bool = False,
                 check_response: Callable[[httpx.Response],
                                          ResponseValue] = None,
                 after_response: Callable = None,
                 **kwargs) -> None:
        self.url = url
        self.url_map = url_map
        self.value = value
        self.method = method
        self.headers = headers
        self.is_json = is_json
        self.check_response = check_response
        self.after_response = after_response

        # 支持更多数据
        self.kwargs = kwargs
        self.check_validity()

    def check_validity(self):
        if None in (self.method, self.url):
            raise ValueError(
                f'{self.__class__.__name__} [url={self.url}, method={self.method}] is necessary.'
            )

    @property
    def request_args(self) -> dict:
        """ 获取 httpx.request 的请求参数。
        
        注意，当进行 json 请求的时候，不要使用 api 提供的 httpx.post(json=...)，
        而是使用 httpx.post(content=..., headers={'Content-Type': 'application/json'})。
        因为其默认使用 ensure_ascii=True，会导致中文变成 unicode 编码。
        见： httpx._content.encode_json 被 encode_request 调用。注意，requests 库亦如此。
        在这种情况下，微信 API 会报错：
        {'errcode': 40033, 'errmsg': 'invalid charset. please check your request, if include \\uxxxx will create fail! hint: [I1T390898vr29]', 'error': True, 'code': 502, 'message': '40033(invalid charset. please check your request, if include \\uxxxx will create fail! hint: [I1T390898vr29])'}

        原因可能如下，转换成 ascii 码之后，实际的 byte 是不同的。
        
        >>> name = 'zrong曾嵘'
        >>> json.dumps(name, ensure_ascii=False).encode('utf-8')
        >>> b'"zrong\xe6\x9b\xbe\xe5\xb5\x98"'
        >>> json.dumps(name, ensure_ascii=True).encode('utf-8')
        >>> b'"zrong\\u66fe\\u5d58"'
        >>> json.dumps(name, ensure_ascii=True)
        >>> '"zrong\\u66fe\\u5d58"'
        """
        p = {
            'method': self.method,
            'url': self.url,
            'headers': self.headers,
        }
        if isinstance(self.url_map, dict):
            p['url'] = self.url.format_map(self.url_map)
        if self.method == 'GET':
            p['params'] = self.value
        elif self.is_json:
            headers = {'Content-Type': 'application/json'}
            if self.headers:
                self.headers.update(headers)
            else:
                self.headers = headers
            p['content'] = json.dumps(self.value, ensure_ascii=False)
            p['headers'] = self.headers
        else:
            p['data'] = self.value
        if self.kwargs:
            for k, v in self.kwargs.items():
                p[k] = v
        return p

    def __repr__(self) -> str:
        return f'{self.__class__.__name__} {self.request_args}'


class HTTPxMixIn:
    """ 需要 http 请求功能的类，可以混合这个类，以提供自己的 http 请求功能。
    当然也可以使用继承或者组合的方式使用。支持异步和同步调用。

    比较几种回调的区别：

    - RequestValue.check_response(httpx.Response)
        在 **每次** HTTP 请求成功之后调用。传递的参数是 httpx.Response 对象。
        这里的请求成功代表的是 HTTP 请求成功(http_status=200)，
        而非实际语义上 API 的调用成功（因为可能存在参数错误等等）
        异步和同步调用，均可在 RequestValue 中包含这个回调方法。

        在同步调用时，HTTPxMixIn.get 和 HTTPxMixIn.post 只有一次请求，
        可以通过对同步方法返回的 ResponseValue.httpx_response 对象进行处理代替对 httpx.Response 的处理。
        因此在 HTTPxMixIn.get 和 HTTPxMixIn.post 中没有包含 check_response 参数传递。

    - HTTPxMixIn *_async(callback) 中的 callback
        在 **每组** 异步 HTTP 请求成功之后调用。传递的参数是 ResponseValue 或者 list[ResponseValue]。
        同步 HTTP 方法不支持 callback，因为直接使用 return 更高效（request_once_sync/request_sync就是这么做的） 。
        注意在使用 request_once_async_wait_run 进行调用的时候（也就是 get_async 和 post_async 采用的方法），
        传递给 callback 的参数是 ResponseValue，因为即使是异步，也只有一次 http 调用。
        使用 request_async_wait_run 进行调用的时候，传递的参数是 list[ResponseValue]。

    - RequestValue.after_response(ResponseValue)
        一般在 HTTPxMixIn *_async(callback) 的 callback 中调用。
        亦可在获得 Respone 之后的任何时刻调用，这取决于使用场景。

        下面是一个场景：

        在 **一组** 异步调用完成之后，可能需要一些对于请求的清理工作。
        check_response 并不适合做这个清理工作，因为它的参数是 httpx.Respone，此时还没有对 http 返回的值做分析。
        after_response 可以在分析完毕之后再处理，它更加灵活，对参数和返回值都不做限制。
        在一组异步请求之后，如果需要分析这一组返回的信息，并于其他信息进行比较，用 after_response 更合适。
    

    如何选择回调方式：

        check_response 是在 **一次 http 请求** 完成后回调，而 callback 是在 **一组 http 请求** 完成之后回调。
        可以根据回调执行的及时性来决定使用哪一种。

        check_response 是在异步线程中执行的。在这里处理，无法与其它处理结果进行比较。
        callback 则返回到主线程执行。因为已经获取了所有的返回，方便整体处理。
        此时在 callback 中调用 after_response 就是一个合理的选择。
    """
    _client_sync: httpx.Client = None
    """ 同步版本的 client。"""

    _request_values: list[RequestValue] = None

    is_async: bool = False

    method: str = None
    """ 公用方法，若请求不提供 method 则使用这个值。"""

    url: str = None
    """ 公用 url，若请求不提供 url 则使用这个值。"""

    url_map: dict = None
    """ 用于替换 url 中的 map 占位符。"""

    headers: dict = None

    def __init__(self,
                 is_async: bool = False,
                 *,
                 url: str = None,
                 method: str = 'GET',
                 url_map: dict = None,
                 headers: dict = None) -> None:
        self.is_async = is_async
        self.url = url
        self.method = method
        self.url_map = url_map
        self.headers = headers
        self._request_values = []

        if not self.is_async:
            self._client_sync = httpx.Client()

    def clear_request_values(self):
        """ 清除保存的请求队列。"""
        self._request_values = []

    def build_request_value(self,
                            rv: RequestValue = None,
                            *,
                            url: str = None,
                            method: str = None,
                            value: dict = None,
                            url_map: dict = None,
                            headers: dict = None,
                            is_json: bool = False,
                            check_response: Callable[[httpx.Response],
                                                     ResponseValue] = None,
                            after_response: Callable = None,
                            insert_to_list: bool = False,
                            **kwargs):
        """ 返回一个 RequestValue 对象。

        :param insert_to_list: 是否插入请求队列，默认值为 true。
        """
        if isinstance(rv, RequestValue):
            # 传入的参数还要再检查一次数据可用性。
            rv.check_validity()
        else:
            method = method or self.method
            url = url or self.url
            url_map = url_map or self.url_map
            headers = headers or self.headers
            rv = RequestValue(url=url,
                              value=value,
                              method=method,
                              url_map=url_map,
                              is_json=is_json,
                              check_response=check_response,
                              after_response=after_response,
                              **kwargs)
        if insert_to_list:
            self._request_values.append(rv)
        return rv

    def request_once_sync(self, rv: RequestValue) -> ResponseValue:
        """ 同步请求一次。"""
        try:
            http_resp = self._client_sync.request(**rv.request_args)
            # 同步请求很少会使用 check_response，更多是接受返回的 ResponseValue，
            # 然后对其 httpx_response 属性进一步处理。
            if rv.check_response is None:
                return ResponseValue(httpx_response=http_resp, request_value=rv)
            else:
                return rv.check_response(http_resp)
        except httpx.HTTPStatusError as e:
            return ResponseValue(message=str(e),
                                 error=True,
                                 code=e.response.status_code,
                                 httpx_response=e.response,
                                 request_value=rv)
        except httpx.HTTPError as e:
            return ResponseValue(message=str(e),
                                 error=True,
                                 code=400,
                                 request_value=rv)
        except Exception as e:
            return ResponseValue(message=str(e),
                                 error=True,
                                 code=500,
                                 request_value=rv)

    def request_sync(self) -> list[ResponseValue]:
        """ 同步请求队列，返回调用结果 list。"""
        resp_data = []
        for rv in self._request_values:
            resp_data.append(self.request_once_sync(rv))
        return resp_data

    def get(self,
            url: str = None,
            value: dict = None,
            *,
            url_map: dict = None,
            headers: dict = None,
            **kwargs) -> ResponseValue:
        """ 发送一次同步 get 请求。"""
        if self.is_async:
            raise ValueError(
                f'{self.__class__.__name__}.get cannot use in async mode!')
        rv = self.build_request_value(url=url,
                                      value=value,
                                      url_map=url_map,
                                      method='GET',
                                      headers=headers,
                                      is_json=False,
                                      insert_to_list=False,
                                      **kwargs)
        return self.request_once_sync(rv)

    def post(self,
             url: str = None,
             value: dict = None,
             *,
             url_map: dict = None,
             headers: dict = None,
             is_json: bool = True,
             **kwargs) -> ResponseValue:
        """ 发送一次同步 post 请求。"""
        if self.is_async:
            raise ValueError(
                f'{self.__class__.__name__}.post cannot use in async mode!')
        rv = self.build_request_value(url=url,
                                      value=value,
                                      url_map=url_map,
                                      headers=headers,
                                      method='POST',
                                      is_json=is_json,
                                      insert_to_list=False,
                                      **kwargs)
        return self.request_once_sync(rv)

    async def request_once_async(self, client: httpx.AsyncClient,
                                 rv: RequestValue) -> ResponseValue:
        """ 异步请求一次。"""
        try:
            http_resp = await client.request(**rv.request_args)
            if rv.check_response is None:
                return ResponseValue(httpx_response=http_resp, request_value=rv)
            else:
                return rv.check_response(http_resp)
        except httpx.HTTPStatusError as e:
            return ResponseValue(message=str(e),
                                 error=True,
                                 code=e.response.status_code,
                                 httpx_response=e.response,
                                 request_value=rv)
        except httpx.HTTPError as e:
            return ResponseValue(message=str(e),
                                 error=True,
                                 code=400,
                                 request_value=rv)
        except Exception as e:
            return ResponseValue(message=str(e),
                                 error=True,
                                 code=300,
                                 request_value=rv)

    async def request_async_wait_run(self,
                                     callback: Callable[[list[ResponseValue]],
                                                        None] = None):
        """ 异步请求队列，使用 callback 来处理一组 http 响应。"""
        async with httpx.AsyncClient() as client:
            tasks = []
            for rv in self._request_values:
                tasks.append(
                    asyncio.ensure_future(self.request_once_async(client, rv)))

            response_values = await asyncio.gather(*tasks)
            if callback:
                callback(response_values)

    async def request_once_async_wait_run(self,
                                          rv: RequestValue,
                                          callback: Callable[[ResponseValue],
                                                             None] = None):
        """ 异步请求一次，使用 callback 来处理这一次 http 响应。"""
        async with httpx.AsyncClient() as client:
            response_value = self.request_once_async(client, rv)
            if callback:
                callback(response_value)

    def request_async(self,
                      callback: Callable[[list[ResponseValue]], None] = None):
        """ 批量调用异步请求，并传递一个 callback 用于处理 http 响应。"""
        asyncio.run(self.request_async_wait_run(callback))

    def post_async(self,
                   url: str = None,
                   value: dict = None,
                   *,
                   url_map: dict = None,
                   headers: dict = None,
                   is_json: bool = True,
                   callback: Callable[[ResponseValue], None] = None,
                   **kwargs):
        """ 发送一次异步 post 请求。"""
        if not self.is_async:
            raise ValueError(
                f'{self.__class__.__name__}.post_async cannot use in sync mode!'
            )
        rv = self.build_request_value(url=url,
                                      method='POST',
                                      value=value,
                                      url_map=url_map,
                                      headers=headers,
                                      is_json=is_json,
                                      insert_to_list=False,
                                      **kwargs)
        asyncio.run(self.request_once_async_wait_run(rv, callback))

    def get_async(self,
                  url: str = None,
                  value: dict = None,
                  *,
                  url_map: dict = None,
                  headers: dict = None,
                  callback: Callable[[ResponseValue], None] = None,
                  **kwargs):
        """ 发送一次异步 get 请求。"""
        if not self.is_async:
            raise ValueError(
                f'{self.__class__.__name__}.get_async cannot use in sync mode!'
            )
        rv = self.build_request_value(url=url,
                                      method='GET',
                                      value=value,
                                      url_map=url_map,
                                      headers=headers,
                                      is_json=False,
                                      insert_to_list=False,
                                      **kwargs)
        asyncio.run(self.request_once_async_wait_run(rv, callback))



"""
.. _fabik_conf:

fabik.conf
-------------------

提供配置文件支持
"""

import os
import jinja2
from pathlib import Path
from typing import Any, Union
from datetime import datetime
import shutil

import tomllib
import tomli_w
import json
import typer

import fabik
from fabik.error import (
    EnvError,
    echo_error,
    echo_info,
    echo_warning,
    FabikError,
    ConfigError,
    PathError,
)
from fabik.tpl import (
    FABIK_TOML_FILE,
    FABIK_TOML_TPL,
)


FABIK_DATA: str = "root_data"


def merge_dict(x: dict, y: dict, z: dict | None = None) -> dict:
    """合并 x 和 y 两个 dict。

    1. 用 y 的同 key 值覆盖 x 的值。
    2. y 中的新键名 (x 中同级不存在) 增加到 x 中。

    返回一个新的 dict，不修改 x 和 y。

    :param x: x 被 y 覆盖。
    :param y: y 覆盖 x。
    :return: dict
    """
    if z is None:
        z = {}
    # 以 x 的键名为标准，用 y 中包含的 x 键名覆盖 x 中的值
    for xk, xv in x.items():
        yv = y.get(xk, None)
        newv = None
        if isinstance(xv, dict):
            newv = xv.copy()
            # 对于 dict 执行递归替换
            if isinstance(yv, dict):
                z[xk] = {}
                newv = merge_dict(newv, yv, z[xk])
            # 对于 list 直接进行浅复制
            elif isinstance(yv, list):
                newv = yv.copy()
            # 对于标量值（非 None）则直接替换
            elif yv is not None:
                newv = yv
        else:
            newv = xv.copy() if isinstance(xv, list) else xv
            if isinstance(yv, dict) or isinstance(yv, list):
                newv = yv.copy()
            elif yv is not None:
                newv = yv
        z[xk] = newv

    # 将 y 中有但 x 中没有的键加入 z
    for yk, yv in y.items():
        if x.get(yk, None) is None:
            z[yk] = yv
    return z


class FabikConfig:
    """解析和处理 FABIK_TOML 配置文件。"""

    root_data: dict[str, Any] | None = None
    """ 保存 FABIK_TOML 中载入的配置。"""

    __work_dir: Path | None = None
    """ 保存工作文件夹所在路径。"""

    fabik_toml: Path | None = None
    """ 保存 FABIK_TOML 文件路径。"""

    def __init__(
        self,
        work_dir: Path | str | None = None,
        cfg: dict | Path | str | None = None,
    ):
        """初始化全局文件。

        :param Path work_dir: 工作文件夹
        :param Path cofig_file: 配置文件路径
        :param str|dict cfg: 相对于工作文件夹的配置文件地址，或者配置内容本身
        """
        self.__work_dir = None if work_dir is None else Path(work_dir)

        if isinstance(cfg, dict):
            self.root_data = cfg
        else:
            self.fabik_toml = FabikConfig.gen_fabik_toml(
                work_dir=work_dir, conf_file=cfg
            )

    @staticmethod
    def gen_work_dir(work_dir: Path | str | None = None) -> Path:
        """生成工作文件夹。"""
        work_dir = Path.cwd() if work_dir is None else Path(work_dir)
        if not work_dir.is_absolute():
            raise PathError(
                err_type=ValueError(), err_msg=f"{work_dir} is not a absolute path."
            )
        return work_dir

    @staticmethod
    def gen_fabik_toml(
        *, work_dir: Path | str | None = None, conf_file: Path | str | None = None
    ) -> Path:
        """生成 fabik.toml 配置文件路径。"""
        work_dir = FabikConfig.gen_work_dir(work_dir)

        if conf_file is None:
            conf_file = FABIK_TOML_FILE
        if isinstance(conf_file, str):
            conf_file = Path(conf_file)
        if not conf_file.is_absolute():
            conf_file = work_dir.joinpath(conf_file)
        return conf_file.resolve().absolute()

    @property
    def file_exists(self) -> bool:
        """返回配置文件 DUEL_TOML 是否存在。"""
        return self.fabik_toml is not None and self.fabik_toml.exists()

    def load_root_data(self) -> dict[str, Any]:
        """获取主配置文件 duel.toml 并进行简单的检测"""
        if self.root_data is None:
            try:
                if self.fabik_toml is None:
                    raise ConfigError(
                        err_type=FileNotFoundError(),
                        err_msg=f"{FABIK_TOML_FILE} not found.",
                    )
                fabik_toml_file = self.fabik_toml.resolve().absolute()
                self.root_data = tomllib.loads(
                    self.fabik_toml.read_text(encoding="utf-8")
                )
            except FileNotFoundError as e:
                raise ConfigError(err_type=e, err_msg=f"{fabik_toml_file} not found.")
            except tomllib.TOMLDecodeError as e:
                raise ConfigError(
                    err_type=e,
                    err_msg=f"Decode {fabik_toml_file} error: {e}",
                )
        return self.root_data

    def merge_root_data(self, data: Union["FabikConfig", dict]) -> dict:
        """使用提供的数据更新 root_data 中的数据，返回的值用于覆盖当前配置文件。"""
        if not self.file_exists:
            raise ConfigError(
                err_type=FileNotFoundError(), err_msg=f"{self.fabik_toml} not found."
            )
        if isinstance(data, dict):
            return merge_dict(data, self.root_data or {})
        return merge_dict(data.root_data or {}, self.root_data or {})

    def getdir(self, *args, work_dir: Path | None = None) -> Path:
        """基于当前项目的运行文件夹，返回一个 pathlib.Path 对象
        如果传递 basedir，就基于这个 basedir 创建路径

        :param args: 传递路径
        :param Path work_dir: 工作文件夹
        """
        if work_dir is not None:
            return Path(work_dir, *args)
        if self.__work_dir is None:
            raise PathError(err_type=ValueError(), err_msg="Please set work_dir first!")
        return Path(self.__work_dir, *args)

    def getcfg(
        self, *args, default_value: Any = None, data: str | dict = FABIK_DATA
    ) -> Any:
        """递归获取 conf 中的值。getcfg 不仅可用于读取 config.toml 的值，还可以通过传递 data 用于读取任何字典的值。

        :param args: 需要读取的参数，支持多级调用，若级不存在，不会报错。
        :param default_value: 找不到这个键就提供一个默认值。
        :param data: 提供一个 dict，否则使用 cfg_data。
        :return: 获取的配置值
        """
        if data == FABIK_DATA:
            data = self.root_data or {}
        if args and isinstance(data, dict):
            cur_data = data.get(args[0], default_value)
            return self.getcfg(*args[1:], default_value=default_value, data=cur_data)
        return data

    def setcfg(self, *args, value: Any, data: str | dict = FABIK_DATA) -> None:
        """递归设置 conf 中的值。setcfg 不仅可用于设置 config.toml 的值，还可以通过传递 data 用于读取任何字典的值。

        :param args: 需要设置的参数，支持多级调用，若级不存在，会自动创建一个内缪的 dict。
        :param data: 提供一个 dict，否则使用 cfg_data。
        :param value: 需要设置的值。
        """
        if data == FABIK_DATA:
            data = self.root_data or {}
        if args and isinstance(data, dict):
            arg0 = args[0]
            if len(args) > 1:
                cur_data = data.get(arg0)
                if cur_data is None:
                    cur_data = {}
                    data[arg0] = cur_data
                self.setcfg(*args[1:], value=value, data=cur_data)
            else:
                data[arg0] = value
                
    def __repr__(self) -> str:
        return f"""{self.__class__.__name__}
        (work_dir={self.__work_dir!s}, 
        fabik_toml={self.fabik_toml!s}, 
        root_data={self.root_data!s})"""


class ConfigWriter:
    """写入配置文件。"""

    tpl_name: str
    dst_file: Path
    replace_obj: dict

    def __init__(
        self,
        tpl_name: str,
        dst_file: Path,
        replace_obj: dict,
    ) -> None:
        """初始化
        :param tplname: 模版名称，不含扩展名
        :param dstname: 目标名称
        """
        self.tpl_name = tpl_name
        self.dst_file = dst_file
        self.replace_obj = replace_obj

    def _fill_meta_data(self, config_data: dict) -> dict:
        """向配置填充 meta 信息。"""
        config_data["create_time"] = f"{fabik.__now__!s}"
        config_data["fabik_version"] = f"fabik v{fabik.__version__}"
        return config_data

    def _write_key_value(self):
        """输出 key = value 形式的文件"""
        txt = "\n".join([f"{k} = {v}" for k, v in self.replace_obj.items()])
        self.dst_file.write_text(txt)
        echo_info(f"文件 {self.dst_file.as_posix()} 创建成功。")

    def _write_file_by_type(self):
        """根据后缀决定如何处理渲染。"""
        if self.tpl_name.endswith(".toml"):
            self.dst_file.write_text(tomli_w.dumps(self.replace_obj))
        elif self.tpl_name == ".env":
            self._write_key_value()
        else:
            # 对于不支持的文件类型，使用 json 格式渲染
            self.dst_file.write_text(
                json.dumps(self.replace_obj, ensure_ascii=False, indent=4)
            )

    def write_file(self, force: bool = True):
        """写入配置文件
        :param force: 若 force 为 False，则仅当文件不存在的时候才写入。
        """
        if not self.dst_file.exists():
            self._write_file_by_type()
        elif force:
            bak_file = self.dst_file.parent.joinpath(
                f"{self.dst_file.name}.bak_{int(fabik.__now__.timestamp())}"
            )
            # 强制覆盖的时候备份原始文件
            shutil.copyfile(self.dst_file, bak_file)
            echo_warning(
                f"备份文件 {self.dst_file.as_posix()} 到 {bak_file.as_posix()}。"
            )
            self._write_file_by_type()
        else:
            echo_error(
                f"文件 {self.dst_file.as_posix()} 已存在。可使用 --force 强制覆盖。"
            )


class TplWriter(ConfigWriter):
    """基于模板文件的配置写入器"""

    tpl_filename: str
    tpl_dir: Path
    tpl_env: jinja2.Environment

    def __init__(
        self,
        tpl_name: str,
        dst_file: Path,
        replace_obj: dict[str, Any],
        tpl_dir: Path,
    ) -> None:
        """初始化
        :param tplname: 模版名称，不含扩展名
        :param dstname: 目标名称
        :param tpl_dir: 模板文件目录
        """
        super().__init__(tpl_name, dst_file, replace_obj)

        # 自动增加 jinja2 后缀，模板文件必须带有 jinja2 后缀。
        self.tpl_filename = (
            tpl_name if tpl_name.endswith(".jinja2") else f"{tpl_name}.jinja2"
        )
        self.tpl_dir = tpl_dir
        self.tpl_env = jinja2.Environment(loader=jinja2.FileSystemLoader(self.tpl_dir))

    def _write_by_jinja(self):
        """能找到模板文件，调用 jinja2 直接渲染。"""
        tpl = self.tpl_env.get_template(self.tpl_filename)
        self.dst_file.write_text(tpl.render(self.replace_obj))
        echo_info(
            f"从模板 {self.tpl_filename} 创建文件{self.dst_file.as_posix()} 成功。"
        )

    def _write_file_by_type(self):
        """重写父类的方法，仅支持 jinja2 模版渲染。"""
        self._write_by_jinja()


class ConfigReplacer:
    env_name: str | None = None
    """ 不提供 env_name，就不会读取 ENV 中的值。"""

    envs: dict | None = None
    fabik_name: str
    fabik_conf: dict
    work_dir: Path
    tpl_dir: Path | None = None
    deploy_dir: Path | None = None
    replace_environ: list[str] | None = None
    writer: ConfigWriter | None = None

    def __init__(
        self,
        fabik_conf: dict[str, Any],
        work_dir: Path,
        tpl_dir: Path | None = None,
        env_name: str | None = None,
    ):
        """初始化"""
        self.env_name = env_name
        self.fabik_conf = fabik_conf
        self.envs = fabik_conf.get("ENV", None)
        self.work_dir = work_dir
        self.tpl_dir = tpl_dir

        self.check_env_name()
        self._set_replace_keys()

    def check_env_name(self):
        """仅在 env_name 有效时，才会检查 envs 的值。"""
        if self.env_name:
            if not self.envs or not isinstance(self.envs, dict):
                raise EnvError(err_type=TypeError(), err_msg="envs must be a dict.")
            if self.env_name not in self.envs:
                raise EnvError(
                    err_type=KeyError(),
                    err_msg=f"env_name {self.env_name} not found in envs.",
                )

    def get_env_value(self, key: str | None = None, default_value: Any = None) -> Any:
        """获取环境变量中的值。"""
        if self.env_name is None or self.envs is None:
            return default_value
        env_obj = self.envs.get(self.env_name, None)
        if env_obj is None:
            return default_value
        if key is None:
            return env_obj
        return env_obj.get(key, default_value)

    def _set_replace_keys(self):
        """设置替换用的 key"""
        self.fabik_name = self.fabik_conf.get("NAME", "fabik")
        self.deploy_dir = Path(self.fabik_conf.get("DEPLOY_DIR", "/srv/app"))
        if not self.deploy_dir.is_absolute():
            raise ConfigError(
                err_type=ValueError(),
                err_msg=f"DEPLOY_DIR must be a absolute path! Now is {self.deploy_dir}.",
            )
        self.replace_environ = self.get_tpl_value("REPLACE_ENVIRON", merge=False)

    def get_tpl_value(
        self, tpl_name: str, merge: bool = True, wrap_key: str | None = None
    ) -> Any:
        """获取配置模版中的值。
        :param tpl_name: 配置模版的键名
        :param merge: 是否合并，对于已知的标量，应该选择不合并
        :param wrap_key: 是否做一个包装。如果提供，则会将提供的值作为 key 名，在最终值之上再包装一层
        """
        base_obj = self.fabik_conf.get(tpl_name, None)
        update_obj = self.get_env_value(tpl_name)
        
        # 对于非字典类型，强制不使用合并
        if not isinstance(base_obj, dict) or not isinstance(update_obj, dict):
            merge = False
            
        if merge:
            repl_obj = merge_dict(base_obj or {}, update_obj or {})
        else:
            repl_obj = update_obj or base_obj
            
        return {wrap_key: repl_obj} if wrap_key else repl_obj

    def replace(self, value: str) -> str:
        """替换 value 中的占位符"""
        # 环境变量替换用
        environ_keys = {}
        # 替换 {NAME} 和 {WORK_DIR}
        replace_obj = {
            "NAME": self.fabik_name,
            "WORK_DIR": self.work_dir.resolve().as_posix(),
        }
        # 增加 {DEPLOY_DIR} 的值进行替换
        if isinstance(self.deploy_dir, Path):
            replace_obj["DEPLOY_DIR"] = self.deploy_dir.as_posix()
        # 获取环境变量中的替换值
        if self.replace_environ is not None:
            for n in self.replace_environ:
                # FABIK_LOCAL_NAME
                environ_key = (
                    f"{self.fabik_name.upper()}_{self.env_name.upper()}_{n}"
                    if self.env_name
                    else f"{self.fabik_name.upper()}_{n}"
                )
                environ_keys[n] = environ_key
                environ_value = os.environ.get(environ_key)
                if environ_value is not None:
                    replace_obj[n] = environ_value
        try:
            # print(f'replace format_map {value=} {replace_obj=}')
            templ: jinja2.Template = jinja2.Template(value)
            # new_value = value.format_map(replace_obj)
            new_value = templ.render(replace_obj)
            return new_value
        except KeyError as e:
            # 抛出对应的 environ key 的错误
            error_key = e.args[0]
            raise FabikError(
                e,
                err_type=e,
                err_msg=f"""error_key: {error_key}
environ_keys: {environ_keys}
replace_obj: {replace_obj}.""",
            )

    def get_replace_obj(self, tpl_name: str) -> dict:
        """获取已经替换过所有值的对象。"""
        replace_obj = self.get_tpl_value(tpl_name)
        replace_str = tomli_w.dumps(replace_obj)
        # 将 obj 转换成 toml 字符串，进行一次替换，然后再转换回 obj
        # 采用这样的方法可以不必处理复杂的层级关系
        replace_obj = tomllib.loads(self.replace(replace_str))
        return replace_obj

    def set_writer(
        self,
        tpl_name: str,
        force=True,
        target_postfix: str = "",
        immediately: bool = False,
    ) -> tuple[Path, Path]:
        """写入配置文件"""
        replace_obj = self.get_replace_obj(tpl_name)
        # 不加后缀的文件路径
        target = self.work_dir.joinpath(tpl_name)
        # 加入后缀的文件路径，大部分情况下与 target 相同
        final_target = self.work_dir.joinpath(f"{tpl_name}{target_postfix}")

        # 基于是否提供了 tpl_dir 决定使用哪种写入器
        self.writer = (
            ConfigWriter(tpl_name, final_target, replace_obj)
            if self.tpl_dir is None
            else TplWriter(tpl_name, final_target, replace_obj, self.tpl_dir)
        )

        if immediately:
            self.writer.write_file(force)
        return target, final_target


def check_none(value: Any, key_list: list[str]) -> Any:
    if value is None:
        echo_error(f'Key "{key_list!s}" not found in config.')
        raise typer.Abort()
    return value


def check_path_exists(p: Path) -> bool:
    if not p.is_absolute():
        echo_error(f"{p!s} is not a absolute path.")
        raise typer.Exit()
    if not p.exists():
        echo_error(f"{p!s} is not a exists path.")
        raise typer.Exit()
    return True


# 默认验证器函数
def config_validator_name_workdir(config: FabikConfig) -> bool:
    """检查 NAME 和 work_dir 是否存在"""
    keys = (["NAME"], ["PATH", "work_dir"])
    for key_list in keys:
        value = check_none(config.getcfg(*key_list), key_list)
        if key_list[-1] == "work_dir":
            check_path_exists(Path(value))
    return True


def config_validator_tpldir(config: FabikConfig) -> bool:
    """默认的配置验证逻辑，检查必要的配置项是否存在"""
    key_list = ["PATH", "tpl_dir"]
    value = check_none(config.getcfg(*key_list), key_list)
    check_path_exists(Path(value))
    return True



""" .. _fabik_error:

fabik.error
----------------------

错误处理。
"""

from typing import Any
from rich.console import Console
from rich.style import Style
from rich.panel import Panel

console: Console = Console(highlight=True)

class FabikError(Exception):
    err_type: Exception
    err_msg: str

    def __init__(self, *args: object, err_type: Exception, err_msg: str) -> None:
        super().__init__(*args)
        self.err_type = err_type
        self.err_msg = err_msg

    def __repr__(self) -> str:
        return f'{self.err_type.__class__.__name__} {self.err_msg} {self.args}'
    

class EncryptError(FabikError):
    """ 加密和解密错误。"""
    pass


class ConfigError(FabikError):
    """ 配置文件错误。"""
    pass


class PathError(FabikError):
    """ 路径错误。"""
    pass

class EnvError(FabikError):
    """ 环境设置错误。"""
    pass


def echo(value:Any, *, panel_title:str | None=None, style:Style | str | None=None):
    """  调用 console.print 输出信息。"""
    if panel_title:
        value = Panel(value, title=panel_title, title_align='left')
    console.print(value, style=style)
    

def echo_info(value: Any, *, panel_title:str | None=None):
    """ 输出信息。"""
    echo(value, panel_title=panel_title)


def echo_warning(value: Any, *, panel_title:str | None=None):
    """ 输出警告。"""
    echo(value, panel_title=panel_title, style='yellow')


def echo_error(value: Any, *, panel_title:str | None=None):
    """ 输出错误。"""
    echo(value, panel_title=panel_title, style='red')
    



import sys
import os

if __name__ == '__main__':
    if not __package__:
        path = os.path.join(os.path.dirname(__file__), os.pardir)
        sys.path.insert(0, path)
    import fabik.cli
    fabik.cli.main()


"""
.. _fabik_log:

fabik.log
-------------------

提供 logging 支持
"""

import sys
import logging
from logging.handlers import WatchedFileHandler
from logging import StreamHandler, Formatter
from pathlib import Path

TEXT_LOG_FORMAT = """
[%(asctime)s] %(levelname)s in %(module)s.%(funcName)s [%(pathname)s:%(lineno)d]:
%(threadName)s:%(processName)s:%(created)s
%(message)s"""


def _create_file_handler(target, filename, chmod=False):
    """创建一个基于文件的 log handler
    :param target: 一个 Path 对象，或者一个 path 字符串
    """
    logsdir = None
    if isinstance(target, Path):
        logsdir = target
    else:
        logsdir = Path(target)
    # 如果是创建文件夹，权限就是 777，如果文件夹存在，又没有调用 chmod，则使用原来的权限
    # 创建或者设置 logs 文件夹的权限，让其他 user 也可以写入（例如nginx）
    # 注意，要设置 777 权限，需要使用 0o40777 或者先设置 os.umask(0)
    # 0o40777 是根据 os.stat() 获取到的 st_mode 得到的
    if not logsdir.exists():
        logsdir.mkdir(mode=0o40777)
    if chmod:
        logsdir.chmod(0o40777)
    logfile = logsdir.joinpath(filename + '.log')
    if not logfile.exists():
        logfile.touch()
    # 使用 WatchedFileHandler 在文件改变的时候自动打开新的流，配合 logrotate 使用
    return WatchedFileHandler(logfile, encoding='utf8')


def get_logger(
    name: str,
    target: Path | str,
    type_: str = 'file',
    fmt: str = 'text',
    level: int = logging.INFO,
):
    """基于 target 创建一个 logger

    :param name: logger 的名称，不要带扩展名
    :param target: 项目主目录的的 path 字符串或者 Path 对象，
        也可以是 tcp://127.0.0.1:8334 这样的地址
    :param type_: stream/file/zmq/pyzog 若使用 pyzog ，
        则调用 get_pyzog_handler，fmt 参数必须为 config_dict
    :param fmt: ``text/json/config_dict``
        如果 ``type_`` 参数为 pyzog，则必须为 ``config_dict``
    :param level: logging 的 level 级别
    :return: 一个 Logger 对象
    """
    hdr = None
    formatter: Formatter | None = None

    if type_ == 'file':
        hdr = _create_file_handler(target, name)
    else:
        hdr = StreamHandler(sys.stdout)
    if fmt == 'raw':
        formatter = Formatter()
    elif fmt == 'default':
        formatter = Formatter(TEXT_LOG_FORMAT)
    else:
        formatter = Formatter(fmt)

    hdr.setLevel(level)
    hdr.setFormatter(formatter)

    log_ = logging.getLogger(name)
    log_.addHandler(hdr)
    log_.setLevel(level)
    return log_



""" .._fabik_util_date:

fabik.util.date
~~~~~~~~~~~~~~~~~~~

日期相关的小工具

@author zrong
@created 2022-05-08
"""
from datetime import date, datetime, timedelta
from collections.abc import Iterable, Iterator, Callable


_DSI = datetime | str | int


class DateRange(Iterable):
    """根据提供的字符串获取日期列表。

    :param date_text: 月份字符串
        单年：2020
        多年：2019,2020
        年范围：2019-2020
        单月：202012
        多月：202012,202103
        月范围：202009-20210
        单日： 20200724
        多日： 20200724,20210103,20220321
        日范围： 20200724-20220301

    :param last_month: 停止在提供的这个月，仅当提供的是年或者月的时候有效。
    :raise ValueError: 如果月份检测不合格抛出 ValueError。
    """

    DATE_FMT: str = '%Y%m%d'
    MONTH_FMT: str = '%Y%m'

    date_type: str = None
    """ 日期类型，值可以是 year/month/date。"""

    date_text: str = None
    """ 输入的日期字符串。"""

    last_month: int = None
    """ 当提供的是年数据时，默认行为是生成这一年的所有月。但有时需要部分月，提供这个参数可以让月份停止在 last_month。"""

    is_range: bool = False
    """ 日期是否为 range 类型，输入字符串中使用 - 分隔两个日期的为 range 类型。"""

    parsed_date: list[str] = None

    def __init__(self, date_text: str, last_month: int = None) -> None:
        if not isinstance(date_text, str) or not date_text.strip():
            raise ValueError(f'提供的日期 {date_text} 不正确！')

        self.date_text = date_text
        self.last_month = last_month

        if date_text.find(',') > 0:
            # 删除重复值，int 为了确保日期是合法整数，str 是方便后面使用 strptime
            parsed_date: list[str] = self.__check_dupli(date_text.split(','))
            self.parsed_date = parsed_date
            self.is_range = False
        # 提供的是范围
        elif date_text.find('-') > 0:
            # 删除重复值，int 为了确保日期是合法整数，str 是方便后面使用 strptime
            parsed_date: list[str] = self.__check_dupli(date_text.split('-'))
            # 不允许的范围表达
            if len(parsed_date) > 2:
                raise ValueError(f'提供的日期 {date_text} 不正确！')
            self.parsed_date = parsed_date
            self.is_range = len(parsed_date) == 2
            if self.is_range:
                if int(self.parsed_date[0]) > int(self.parsed_date[1]):
                    self.parsed_date = [self.parsed_date[1], self.parsed_date[0]]
        else:
            self.is_range = False
            self.parsed_date = [date_text]

        date_len: set[int] = {len(d) for d in self.parsed_date}
        if len(date_len) > 1:
            raise ValueError(f'提供的日期 {date_text} 长度不一致！')

        check_date_len: int = date_len.pop()
        # 使用的是日期
        if check_date_len == 8:
            self.date_type = 'date'
        elif check_date_len == 6:
            self.date_type = 'month'
        elif check_date_len == 4:
            self.date_type = 'year'
        else:
            raise ValueError(f'提供的日期 {date_text} 长度不正确，仅支持 8/6/4 三种日期长度。')

    def __iter__(self) -> Iterator:
        return iter(self.to_list())

    def __len__(self) -> int:
        return len(self.to_list())

    @classmethod
    def make_today(cls, fmt: str = None) -> str:
        """获取今天的字符串形式，若未提供 fmt 参数，则使用 DATE_FMT。"""
        return date.today().strftime(fmt or cls.DATE_FMT)

    @classmethod
    def make_this_year(cls, fmt: str = None) -> str:
        """获取今年的第一天到今天的范围字符串。 若未提供 fmt 参数，则使用 DATE_FMT。"""
        today = date.today()
        first_day_of_this_year = date(today.year, 1, 1).strftime(fmt or cls.DATE_FMT)
        return f'{first_day_of_this_year}-{today.strftime(fmt or cls.DATE_FMT)}'

    def __check_dupli(self, parsed_date: list[str]) -> list[str]:
        """删除重复值，int 为了确保日期是合法整数，str 是方便后面使用 strptime。"""
        return [str(dl) for dl in {int(d) for d in parsed_date}]

    def __range_star_or_end(self, index: int, type_: Callable = str) -> _DSI:
        if not self.is_range:
            return None
        s = self.parsed_date[index]
        if type_ == datetime:
            if self.date_type == 'date':
                return datetime.strptime(s, self.DATE_FMT)
            elif self.date_type == 'month':
                return datetime.strptime(s, self.MONTH_FMT)
            return datetime(int(s), 1, 1)
        return type_(s)

    def range_start(self, type_: Callable = str) -> _DSI:
        return self.__range_star_or_end(0, type_)

    def range_end(self, type_: Callable = str) -> _DSI:
        return self.__range_star_or_end(1, type_)

    def to_list(self) -> list[int]:
        """根据当前 date_type 导出不同的 list。"""
        if self.date_type == 'date':
            return self.to_date_list()
        return self.to_month_list()

    def to_date_list(self) -> list[int]:
        """导出一个 8 位 date 列表。

        :return list[int]: 返回 %Y%m%d 的整数形式列表。
        """
        if self.date_type != 'date':
            raise TypeError('请确保 date_type 为 date！')
        if self.is_range:
            # 小值在前
            start: datetime = self.range_start(datetime)
            end: datetime = self.range_end(datetime)
            date_list: list[int] = []
            # 构建日期列表
            for i in range((end - start).days + 1):
                day = start + timedelta(days=i)
                date_list.append(int(day.strftime(self.DATE_FMT)))
            return date_list
        return [
            int(datetime.strptime(date_str, self.DATE_FMT).strftime(self.DATE_FMT))
            for date_str in self.parsed_date
        ]

    def to_month_list(self) -> list[int]:
        """导出一个 6 位 month 列表。

        :return list[int]: 返回 %Y%m 的整数形式列表。
        """
        if not self.date_type in ('month', 'year'):
            raise TypeError('请确保 date_type 为 month 或 year！')

        year2monthlist = lambda y: [int(f'{y}{m:0>2d}') for m in range(1, 13)]
        month = []
        if self.is_range:
            # 使用的是月份
            if self.date_type == 'month':
                start = self.range_start(datetime)
                end = self.range_end(datetime)
                y_start = start.year
                m_start = start.month
                y_end = end.year
                m_end = end.month
                y = y_start
                while y <= y_end:
                    for m in range(1, 13):
                        # 开始年份要判断 y_start
                        if y == y_start and m < m_start:
                            continue
                        # 结束年份要判断 y_end
                        if y == y_end and m > m_end:
                            continue
                        month.append(int(f'{y}{m:0>2d}'))
                    y += 1
            # 使用的年份
            else:
                y_start = int(self.parsed_date[0])
                y_end = int(self.parsed_date[1])
                y = y_start
                while y <= y_end:
                    month.extend(year2monthlist(y))
                    y += 1
        # 提供的是列表
        else:
            # 使用的是月份
            if self.date_type == 'month':
                month = [int(m) for m in self.parsed_date]
            # 使用的年份
            else:
                for y in self.parsed_date:
                    month.extend(year2monthlist(y))
        # 检测月份是否正规
        [datetime.strptime(str(m), self.MONTH_FMT) for m in month]
        if self.last_month is not None:
            return [m for m in month if m <= self.last_month]
        return month

    def to_year_list(self) -> list[int]:
        """导出一个 4 为的 year 列表。

        :return list[int]: 返回 %Y 的整数形式列表。
        """
        month_list = self.to_month_list()
        year_list = [int(month / 100) for month in month_list]
        # 去重
        return list(set(year_list))


def date_interval(date1: _DSI, date2: _DSI, fmt: str = '%Y%m%d') -> int:
    d1: datetime = (
        date1 if isinstance(date1, datetime) else datetime.strptime(str(date1), fmt)
    )
    d2: datetime = (
        date2 if isinstance(date2, datetime) else datetime.strptime(str(date2), fmt)
    )
    d: timedelta = d1 - d2
    return abs(d.days)


def get_years(start: int = 2017, end: int = None) -> list[int]:
    """获取年份列表。

    :param start: 开始的年份。
    :param end: 结束的年份。若值为 None 代表今年。
    """
    if end is None:
        now = date.today()
        return range(start, now.year + 1)
    return range(start, end + 1)


def get_last_month(month: int = None) -> int:
    """获取上一个月的 %Y%m 表达。

    :param month: 获取 month 的上一个月。若 month 为 None，则使用 today。
    """
    if month is None:
        today = date.today()
    else:
        today = datetime.strptime(str(month), '%Y%m')
    last_month = today - timedelta(days=today.day)
    return int(last_month.strftime('%Y%m'))


def get_last_12month(month: int = None) -> str:
    """获取最近 12 个月的 %Y%m 表达，形如 202104-202203。

    :param month: 获取 month 的上一个月。若 month 为 None，则使用 today。
    """
    year_month = get_last_month(month)
    dt = datetime.strptime(str(year_month), '%Y%m')
    last_12month = dt - timedelta(days=365)
    return last_12month.strftime('%Y%m') + '-' + dt.strftime('%Y%m')


def gen_month(year: int, start: int = 1, end: int = 12, type_: Callable = int) -> list:
    """生成月份列表。"""
    return [type_(f'{year}{md:02}') for md in range(start, end + 1)]


def from_month(
    month_text: str, last_month: int = None, use_year: bool = False
) -> list[int]:
    """根据提供的字符串获取日期列表。

    :param month_text: 月份字符串
        单年：2020
        多年：2019,2020
        年范围：2019-2020
        单月：202012
        多月：202012,202103
        月范围：202009-20210
    :param last_month: 停止在提供的这个月。
    :param year_list: 返回年份列表而不是月份列表。
    :raise ValueError: 如果月份检测不合格抛出 ValueError。
    :return list[int]: 返回 %Y%m 的整数形式列表。
    """
    dr = DateRange(month_text, last_month)
    if use_year:
        return dr.to_year_list()
    return dr.to_month_list()


def month2date(month_text: str, last: bool = False) -> int:
    """将一个 month_text 转换成一个日期。
    如果提供的年，则转换为当年的一个日期。
    如果提供的是月，则转换为当月的一个日期。
    若提供的是日期，则直接转换为整数。

    :param month_text: 见 from_month。
    :param last: 若为 True 转换为最后日期，否则转换为首日。
    """
    if len(month_text) == 8:
        return int(month_text)
    mlist = from_month(month_text)
    m = mlist[-1] if last else mlist[0]
    # 找到该月的最后一天
    if last:
        next_month = date(int(str(m)[:4]), int(str(m)[4:6]), 28) + timedelta(days=4)
        d = next_month - timedelta(days=next_month.day)
        return int(d.strftime('%Y%m%d'))
    return int(f'{m}01')



""" .. _fabik_deploy_tmux:

fabik.deploy.tmux
~~~~~~~~~~~~~~~~~~~

封装远程 tmux 调用。
"""

class Tmux(object):
    """Tmux helper for fabric > 2"""
    def __init__(self, runner, session_name='default'):
        self.session_name = session_name
        self.run_cmd = runner.run

        self.create_session()

    def create_session(self):
        test = self.run_cmd('tmux has-session -t %s' % self.session_name, warn=True)

        if test.failed:
            self.run_cmd('tmux new-session -d -s %s' % self.session_name)

        self.run_cmd(
            'tmux set-option -t %s -g allow-rename off' % self.session_name)

    def recreate(self):
        self.kill_session()
        self.create_session()

    def kill_session(self):
        self.run_cmd('tmux kill-session -t %s' % self.session_name)

    def command(self, command, pane=0):
        self.run_cmd('tmux send-keys -t %s:%s "%s" ENTER' % (
            self.session_name, pane, command))

    def new_window(self, name):
        self.run_cmd('tmux new-window -t %s -n %s' % (self.session_name, name))

    def find_window(self, name):
        test = self.run_cmd('tmux list-windows -t %s | grep \'%s\'' % (self.session_name, name), warn=True)

        return test.ok

    def rename_window(self, new_name, old_name=None):
        if old_name is None:
            self.run_cmd('tmux rename-window %s' % new_name)
        else:
            self.run_cmd('tmux rename-window -t %s %s' % (old_name, new_name))

    def wait_for(self, signal_name):
        self.run_cmd('tmux wait-for %s' % signal_name)

    def run_singleton(self, command, orig_name, wait=True):
        run_name = "run/%s" % orig_name
        done_name = "done/%s" % orig_name

        # If the program is running we wait to be finished.
        if self.find_window(run_name):
            self.wait_for(run_name)

        # If the program is not running we create a window with done_name
        if not self.find_window(done_name):
            self.new_window(done_name)

        self.rename_window(run_name, done_name)

        # Check that we can execute the commands in the correct window
        assert self.find_window(run_name)

        rename_window_cmd = 'tmux rename-window -t %s %s' % (
            run_name, done_name)
        signal_cmd = 'tmux wait-for -S %s' % run_name

        expanded_command = '%s ; %s ; %s' % (
            command, rename_window_cmd, signal_cmd)
        self.command(expanded_command, run_name)

        if wait:
            self.wait_for(run_name)



""" .. _fabik_deploy_gunicorn:

fabik.deploy.gunicorn
~~~~~~~~~~~~~~~~~~~~~~

封装远程 gunicorn 部署。
"""

from pathlib import Path
from fabric.connection import Connection
from invoke.exceptions import Exit

from fabik.deploy import Deploy, logger


class GunicornDeploy(Deploy):
    """ 使用 Gunicorn 来部署服务
    """
    def __init__(self, fabik_conf: dict, work_dir: Path, conn: Connection, env_name: str | None=None):
        super().__init__(fabik_conf, work_dir, conn, env_name)

    def get_pid_file(self):
        """ 使用 pidfile 来判断进程是否启动
        """
        pidfile = self.get_remote_path('gunicorn.pid')
        if self.remote_exists(pidfile):
            return pidfile
        return None

    def get_pid_value(self, *args):
        """ 获取远程 pid 文件中的 pid 值
        """
        pid_value = self.cat_remote_file('gunicorn.pid')
        if pid_value is None:
            raise Exit('gunicorn.pid 没有值！')
        return pid_value.strip()

    def get_gunicorn_exe(self):
        """ 获取 venv 中 uwsgi 的可执行文件绝对路径
        """
        gunicorn_exe = self.get_remote_path('venv/bin/gunicorn')
        if not self.remote_exists(gunicorn_exe):
            raise Exit('没有找到 gunicorn 可执行文件！请先执行 init_remote_venv')
        return gunicorn_exe

    def start(self, wsgi_app=None, daemon=None):
        """ 启动服务进程
        :@param wsgi_app: 传递 wsgi_app 名称
        :@param daemon: 若值为 True，则强制加上 -D 参数
        """
        pidfile = self.get_pid_file()
        if pidfile is not None:
            raise Exit('进程不能重复启动！')
        conf = self.get_remote_path('gunicorn.conf.py')
        cmd = self.get_gunicorn_exe() + ' --config ' + conf
        if daemon is True:
            cmd += ' -D'
        if wsgi_app is not None:
            cmd += ' ' + wsgi_app
        self.conn.run(cmd)

    def stop(self):
        """ 停止 API 进程
        """
        pidvalue = self.get_pid_value()
        killr = self.conn.run('kill -s TERM ' + pidvalue)
        if killr.ok:
            logger.warning('优雅关闭 %s', pidvalue)
            # 删除 pidfile 以便下次启动
            self.conn.run('rm %s' % self.get_pid_file())
        else:
            logger.warning('关闭 %s 失败', pidvalue)

    def reload(self):
        """ 优雅重载 API 进程
        """
        pidvalue = self.get_pid_value()
        killr = self.conn.run('kill -s HUP ' + pidvalue)
        if killr.ok:
            logger.warning('优雅重载 %s', pidvalue)
        else:
            logger.warning('重载 %s 失败', pidvalue)



""" .._fabik_util:

fabik.util
------------------

封装的小工具。
"""

import hashlib
import base64


def md5txt(txt: str | bytes) -> str:
    """
    计算 MD5 字符串散列
    :param txt:
    :return:
    """
    md5obj = hashlib.md5()
    if isinstance(txt, str):
        txt = txt.encode('utf-8')
    md5obj.update(txt)
    return md5obj.hexdigest()


def md5base64(txt: str | bytes) -> str:
    """md5(base64) 算法
    验证工具： http://www.cmd5.com/hash.aspx
    """
    m = hashlib.md5()
    if isinstance(txt, str):
        txt = txt.encode('utf-8')
    m.update(txt)
    return base64.encodebytes(m.digest())[:-1].decode('utf8')


# 在函数定义之后导入gen，避免循环导入
from . import gen

__all__ = ['md5txt', 'md5base64', 'gen']
